"""Ce fichier contient une implémentation non déterministe de la classe
abstraite Autoencodeur.
"""
import os  # Utile pour les manipulations de fichiers.
import numpy as np  # Utile pour fournir des données à Keras.

from keras.layers import (
    Input,
    Dense,
    Conv2D,
    MaxPooling2D,
    UpSampling2D,
)  # Les éléments constitutifs de notre modèle.
from keras.models import (
    Model,
    load_model,
)  # Outils pour manipuler des modèles Keras facilement
from keras import backend as K

from intelligenceArtificielle.Entrainement import (
    lectureBlob,
)  # Des fonctions utiles pour simplifier l'entrainement non supervisé
from intelligenceArtificielle.Autoencodeur import (
    Autoencodeur,
)  # La classe que l'on cherche à implémenter


class AutoencodeurNonDeterministe(Autoencodeur):
    """Classe pour les autoencodeurs non déterministes.

    Variables globales
    ------------------
    TAILLE_GROUPE : int
        Le nombre d'images à générer en même temps lors de la recherche d'une
        amélioration pour un autoencodeur non déterministe.
    ECHEC_PERMIS : float
        La proportion de la population pour laquelle on autorise l'autoencodeur
        non déterministe à ne pas trouver d'améliorations. Plus cette proportion
        est grande, plus l'amélioration est rapide mais moins elle est efficace.

    Attributs
    ---------
    largeur : int
        Dimension des images. Utile dans plusieurs méthodes distinctes.
    hauteur : int
        Dimension des images. Utile dans plusieurs méthodes distinctes.
    autoencodeur : modèle Keras
        Modèle choisi.
    """

    TAILLE_GROUPE = 10
    ECHEC_PERMIS = 0.1

    def __init__(
        self,
        nomDuModele="ameliorateur",
        largeur=32,
        hauteur=32,
        *args,
        **kwargs
    ):
        """Initialisation de la classe.

        Paramètres
        ----------
        nomDuModele : str
            Nom du moèle à charger. None si l'on ne souhaite pas charger de
            modèle. On ira charger le fichier "src/modeles/nomDuModele.h5".
        largeur : int
        hauteur : int

        Exceptions levées
        -----------------
        ModeleIntrouvable
            Lorsque nomDuModèle.h5 n'a pas pu être trouvé.
        """
        super(AutoencodeurNonDeterministe, self).__init__(
            nomDuModele, largeur, hauteur, *args, *kwargs
        )
        # Le constructeur est hérité de Autoencodeur, rien de plus à faire ici

    def creation(
        self, facteurReduction=2, baseDimensions=9, baseNoyau=4, baseDense=50
    ):
        """Méthode à utiliser pour créer un nouveau modèle dans le contexte d'un
        entrainement.

        La meilleure configuration d'autoencodeur que j'ai trouvée pour notre
        problème. Il est concu pour des images de taille 32 x 32 mais d'autres
        configurations sont possibles.

        Paramètres
        ----------
        facteurReduction : int
            ou "pool_size". Facteur de réduction pour les convolutions du
            réseau.
        baseDimensions : int | list
            Dimensions des filtres de convolution du réseau.
        baseDense : int | list
            Dimensions pour les filtres "Dense" du réseau.
        baseNoyau : int | list
            ou "kernel_size". Dimensions des noyaux de convolution du réseau.

        Notes
        -----
        On peut donner soit une liste, soit un unique coefficient dans
        `baseDimensions`, `baseDense` et `baseNoyau`. Si on donne une liste, ce
        sont les coefficients dans cette liste qui sont utilisés.
        """

        # Pour baseDimensions
        if type(baseDimensions) is list:
            dimensions = baseDimensions[:4]
        else:
            # On a fourni un unique coefficent en argument
            dimensions = [i * baseDimensions for i in range(1, 5)]

        # De même pour baseNoyau
        if type(baseNoyau) is list:
            noyau = baseNoyau[:4]
        else:
            noyau = [(5 - i) * baseNoyau for i in range(1, 5)]

        # Et enfin pour baseDense
        if type(baseDense) is list:
            dense = baseDense[:2]
        else:
            dense = [2 * baseDense, baseDense]

        # La fonction utilisée par la librairie Keras pour construire notre
        # autoencodeur. Dans la notation Keras, la forme de entree est
        # (None, hauteur, largeur, 3).
        entree = Input(shape=(self.hauteur, self.largeur, 3))

        # x sera de taille (None, hauteur, largeur, 6)
        x = Conv2D(
            filters=dimensions[0],
            kernel_size=noyau[0],
            activation="relu",
            padding="same",
        )(entree)
        # encode sera de taille (None, hauteur/2, largeur/2, 6)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Conv2D(
            filters=dimensions[1],
            kernel_size=noyau[1],
            activation="relu",
            padding="same",
        )(x)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Conv2D(
            filters=dimensions[2],
            kernel_size=noyau[2],
            activation="relu",
            padding="same",
        )(x)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Conv2D(
            filters=dimensions[3],
            kernel_size=noyau[3],
            activation="relu",
            padding="same",
        )(x)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Dense(dense[0], activation="relu")(x)
        # x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Dense(dense[1], activation="relu")(x)

        # x = UpSampling2D(size=facteurReduction)(x)
        x = Dense(dense[0], activation="relu")(x)

        x = UpSampling2D(size=facteurReduction)(x)
        x = Conv2D(
            filters=dimensions[3],
            kernel_size=noyau[3],
            activation="relu",
            padding="same",
        )(x)

        x = UpSampling2D(size=facteurReduction)(x)
        x = Conv2D(
            filters=dimensions[2],
            kernel_size=noyau[2],
            activation="relu",
            padding="same",
        )(x)

        x = UpSampling2D(size=facteurReduction)(x)
        x = Conv2D(
            filters=dimensions[1],
            kernel_size=noyau[1],
            activation="relu",
            padding="same",
        )(x)

        # y sera de taille (None, hauteur, largeur, 6)
        x = UpSampling2D(size=facteurReduction)(x)
        # z sera de taille (None, hauteur, largeur, 6)
        x = Conv2D(
            filters=dimensions[0],
            kernel_size=noyau[0],
            activation="relu",
            padding="same",
        )(x)

        # decode sera de taille (None, hauteur, largeur, 3)
        decode = Conv2D(3, (3, 3), activation="sigmoid", padding="same")(x)

        # On compile le modèle créé.
        self.autoencodeur = Model(entree, decode)
        self.autoencodeur.compile(
            optimizer="adadelta", loss="binary_crossentropy"
        )
        # Utile pour avoir la forme de l'autoencodeur lorsque l'on fait du debuggage.
        self.autoencodeur.summary()

    def ameliorerArbres(self, listeArbres, iterations=1):
        """Version plus haut niveau de `predire` qui travaille directement sur
        des arbres pour plus de simplicité dans l'algorithme génétique.

        Cas non déterministe. Notez que dans le cas non déterministe on suppose
        que si un arbre relit l'image qu'il a lui même créé, il revient dans
        l'état où il était au moment de créer l'image (ce qui est vrai sauf
        problèmes d'arrondi).

        Paramètres
        ----------
        listeArbres : list
        iterations : int
        """

        # On définit la liste des arbres et images pour lesquelles on n'a pas
        # trouvé d'améliorations et qui va diminuer à chaque itération. Notez
        # que l'on conserve aussi l'indice qu'a l'arbre dans la liste initiale
        # afin de renvoyer les images dans le bon ordre à la fin de la fonction
        # (c'est à cela que sert enumerate).
        listeArbreRestants = list(enumerate(listeArbres))
        # Ces images nous servent aussi à restaurer les arbres que nous n'avons
        # pas réussi à améliorer à la fin de la boucle.
        listeImagesAmeliorees = [
            arbre.ecritureImage(largeur=self.largeur, hauteur=self.hauteur)
            for arbre in listeArbres
        ]
        # On a aussi besoin du score initial de tous les arbres
        listeScoresRestants = [arbre.score() for arbre in listeArbres]

        # Le seuil au dela duquel on a améliorer suffisament d'individus pour se
        # permettre de sortir de la boucle.
        seuilTerminaison = (
            len(listeArbres) * AutoencodeurNonDeterministe.ECHEC_PERMIS
        )

        # On boucle tant que l'on a pas trouvé une amélioration pour suffisament
        # d'arbres.
        while len(listeArbreRestants) > seuilTerminaison:
            # Avant de générer des propositions aléatoires, on récupère la liste
            # de toutes les images restant à traiter.
            listeImagesRestantes = []
            for indiceOrigine, arbre in listeArbreRestants:
                listeImagesRestantes.append(
                    listeImagesAmeliorees[indiceOrigine]
                )

            # On génère des propositions aléatoires.
            propositions = [
                self.autoencodeur.predict(np.array(listeImagesRestantes))
                for i in range(AutoencodeurNonDeterministe.TAILLE_GROUPE)
            ]

            for i in range(len(listeArbreRestants)):
                # indiceOrigine est l'indice qu'avait l'arbre dans la liste
                # d'origine.
                indiceOrigine, arbre = listeArbreRestants[i]

                # L'indice de la meilleure proposition parmi toutes celles
                # fournies de façon non deterministe
                meilleureProposition = None
                for j in range(AutoencodeurNonDeterministe.TAILLE_GROUPE):
                    # On lit l'image proposée. Le premier indice correspond au
                    # fait qu'il y a plusieurs propositions par images et le
                    # second au fait qu'il y a plusieurs images.
                    # ATTENTION, lectureImage détruit l'image lue par effet de
                    # bord. On en fait donc une copie afin de ne pas la perdre.
                    copieImage = propositions[j][i].copy()
                    arbre.lectureImage(copieImage)
                    # On calcule le nouveau score
                    scorePropose = arbre.score()
                    if scorePropose > listeScoresRestants[i]:
                        # On note l'index de la proposition
                        meilleureProposition = j
                        # On note aussi le score de la proposition dans la liste
                        # afin de pouvoir y comparer les propositions qui
                        # restent à traiter.
                        listeScoresRestants[i] = scorePropose

                if meilleureProposition is not None:
                    # Si on a trouvé une proposition qui marche bien, on remet
                    # l'arbre à jour et on met l'image correpondante dans la
                    # liste à renvoyer.

                    # Ici nous n'avons pas besoin de faire de copie de l'image,
                    # elle ne sera pas réutilisée de toute façon.
                    meilleureImage = propositions[meilleureProposition][i]
                    arbre.lectureImage(meilleureImage)
                    # On supprime l'individu des individus encore à traiter.
                    del listeArbreRestants[i]
                    del listeScoresRestants[i]
                    # Enfin, on remplace l'image d'origine par l'image améliorée
                    # dans la liste dédiée
                    listeImagesAmeliorees[indiceOrigine] = meilleureImage
                # Notez que nous n'avons rien besoin de faire sinon, l'arbre
                # sera remis à jour correctement en lisant un image lorsque nous
                # testerons de nouvelles propositions.

        # Notez que pour tous les arbres pour lesquels nous n'avons pas trouvé
        # de meilleure proposition, il nous faut les remettre dans leur état
        # d'origine.
        for i in range(len(listeArbreRestants)):
            indiceOrigine, arbre = listeArbreRestants[i]
            imageOrigine = listeImagesAmeliorees[indiceOrigine]
            # On restaure l'arbre à son état initial.
            # ATTENTION, l'image utilisée est détruite par effet de bord.
            arbre.lectureImage(imageOrigine)



# Ce fichier contient l'implémentation des noeuds de type parallèle. La documentation de toutes les méthodes des noeuds est disponible dans la classe parente.

from random import randrange  # Utile pour un élaguage aléatoire

from intelligenceArtificielle.structureDonnees.Noeud import Noeud
from intelligenceArtificielle.structureDonnees.SunRiseException import (
    NonFeuilleException,
    ImageTropPetite,
)


class Parallele(Noeud):
    """docstring for Parallele."""

    def __init__(self, arbre=None, parent=None):
        super(Parallele, self).__init__(arbre=arbre, parent=parent)

    # Note : les propriétés ne sont pas héritées en python...
    @property
    def profondeur(self):
        if self.parent is None:
            return 0
        elif self._profondeur is None:
            self._profondeur = (
                self.parent.profondeur + 1
            )  # On mémoïse la profondeur pour éviter des calculs inutiles
        return self._profondeur

    def creationSimulationRecursive(self, A, B, C, gauche, droite, curseur):
        result = 0
        for fils in self.fils:
            resultBranche, curseur = fils.creationSimulationRecursive(
                A, B, C, gauche, droite, curseur
            )
            if resultBranche is not None:
                result += resultBranche
        return result, curseur

    def creationMarquageRecursif(
        self, A, B, C, D, gauche, droite, curseur, conteneur, capaciteDroite
    ):
        if conteneur is None:
            # Si on n'est pas sous une liaison parallèle, on crée un conteneur.
            nouveauConteneur = []
        else:
            # On réutilise le conteneur du parent
            nouveauConteneur = conteneur

        result = 0
        for fils in self.fils:
            (
                delaiBranche,
                resultBranche,
                curseur,
            ) = fils.creationMarquageRecursif(
                A,
                B,
                C,
                D,
                gauche,
                droite,
                curseur,
                nouveauConteneur,
                capaciteDroite,
            )
            if resultBranche is not None:
                result += resultBranche

        if conteneur is None:
            # i.e. on n'est pas sous une liaison en paralèle, alors on résout le délai des fils.
            delai = sum(nouveauConteneur) / len(nouveauConteneur)
            return delai, result, curseur
        else:
            return None, result, curseur

    def injectionMarquage(self, E, gauche, droite, curseur, capaciteDroite):
        for fils in self.fils:
            curseur = fils.injectionMarquage(
                E, gauche, droite, curseur, capaciteDroite
            )
        return curseur

    def suppressionFils(self, index):
        # On met à jour la forme de l'arbre en détachant le noeud désigné.
        self.fils[index].detacher()
        # IndexError out of range si index est trop grand pour self.fils
        del self.fils[index]
        taille = len(self.fils)
        if taille < 2:
            # S'il ne reste plus qu'un seul bloc fils il prend la place du bloc parallèle. Il faut juste faire attention à la mémoïzation de la profondeur
            remplacant = self.fils[0]
            del self.fils[0]  # Pour éviter les problèmes avec détacher
            remplacant.detacher()  # On va le rattacher dans remplacer
            remplacant._profondeur = self.profondeur
            # Notez que replace appelle detacher pour nous en interne
            self.remplacer(remplacant)
            # On renvoie le noeud qui prend notre place dans l'arbre
            return remplacant
        else:
            return self

    def ajoutFils(self, nouveauFils, index, forme=None):
        if forme is not None:
            raise NonFeuilleException(
                "Vous ne pouvez pas changer le forme d'un noeud (non feuille) qui existe déjà avec ajoutFils !"
            )
        elif index is None:
            raise NonFeuilleException(
                "Vous devez préciser à quel index insérer le nouveau fils parmi les fils du noeud préexistant !"
            )
        if index == len(self.fils):
            self.fils.append(nouveauFils)
        else:
            self.fils.insert(index, nouveauFils)
        # On met à jour la généalogie du noeud ajouté.
        nouveauFils.parent = self
        # On attache le nouveau noeud à l'arbre
        nouveauFils.attacher(self.arbre)
        # On garde notre place dans l'arbre, donc on renvoie self
        return self

    def substituerEnfants(self, listeFils):
        # On commence par détacher tous nos enfants
        for fils in self.fils:
            fils.detacher()
        # Puis on change la liste des noeuds enfants.
        self.fils = listeFils
        for fils in self.fils:
            # On se note comme parent dans chacun des nouveaux fils
            fils.parent = self
            # On attache le nouveau fils à l'arbre
            fils.attacher(self.arbre)

    # L'algorithme de copie du sous arbre est évidemment récursif
    def sousArbre(self):
        copieNoeud = Parallele(None)
        for fils in self.fils:
            copieFils = fils.sousArbre()
            copieFils.parent = copieNoeud
            copieNoeud.fils.append(copieFils)
        return copieNoeud

    def attacher(self, arbre):
        if arbre is not self.arbre:
            self.arbre = arbre
            if self.profondeur + 1 == len(self.arbre.forme):
                # Comme un noeud parallèle a forcément des enfants, il faut ajouter un niveau à la forme de l'arbre
                self.arbre.forme.append(0)
            self.arbre.forme[self.profondeur] += 1
            for fils in self.fils:
                fils.attacher(arbre)

    def detacher(self, perdreParent=True):
        self.arbre.forme[self.profondeur] -= 1
        # On détache récursivement tous les fils pour bien prendre en compte l'influence sur la forme.
        for fils in self.fils:
            fils.detacher(perdreParent=False)
        if self.arbre.forme[self.profondeur + 1] == 0:
            # i.e. on a détaché tous les noeuds qui étaient à la profondeur suivante (qui est par construction la dernière de l'arbre), alors il faut l'enlever de la forme de l'arbre pour éviter les 0 inutiles
            self.arbre.forme.pop()
        # On ne perd la référence à l'arbre qu'à la toute fin de la fonction car on la référence au dessus
        self.arbre = None
        # On perd aussi la mémoïzation de la profondeur pour éviter de la réutiliser si on est à l'intérieur d'un sous arbre qui est déplacé
        self._profondeur = None
        if perdreParent:
            # On perd aussi la référence à son parent pour éviter les effets de bord étranges
            self.parent = None

    def dessiner(self, image, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On calcule la hauteur de chaque subdivision verticale que l'on s'apprète à créer. On rappelle que le 0,0 est tout en haut à gauche de l'image.
        hauteur = int((coinBasDroiteY - coinHautGaucheY) / nombreDivisions)
        if hauteur == 0:
            raise ImageTropPetite(
                f"La hauteur d'une liaison parallèle écrite à la profondeur {self.profondeur} serait 0, la forme de l'arbre est {self.arbre.forme}"
            )
        # On appelle récursivement la méthode dessiner sur les enfants
        for i in range(nombreDivisions - 1):
            fils = self.fils[i]
            coinHautGaucheFils = (
                coinHautGaucheY + i * hauteur,
                coinHautGaucheX,
            )
            coinBasDroiteFils = (
                coinHautGaucheY + (i + 1) * hauteur,
                coinBasDroiteX,
            )
            fils.dessiner(image, coinHautGaucheFils, coinBasDroiteFils)
        # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
        fils = self.fils[-1]
        coinHautGaucheFils = (
            coinHautGaucheY + (nombreDivisions - 1) * hauteur,
            coinHautGaucheX,
        )
        coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
        fils.dessiner(image, coinHautGaucheFils, coinBasDroiteFils)

    # La façon dont on calcule les moyennes pour les capacités dans cette implémentation est sans doute une source d'erreur.
    def lire(self, image, coinHautGauche, coinBasDroite, conteneur):
        if conteneur is None:
            nouveauConteneur = []
        else:
            nouveauConteneur = conteneur
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On calcule la hauteur de chaque subdivision verticale que l'on s'apprète à créer. On rappelle que le 0,0 est tout en haut à gauche de l'image.
        hauteur = int((coinBasDroiteY - coinHautGaucheY) / nombreDivisions)
        if hauteur == 0:
            raise ImageTropPetite(
                f"La hauteur d'une liaison parallèle lue à la profondeur {self.profondeur} serait 0, la forme de l'arbre est {self.arbre.forme}"
            )
        # Dans la mesure où l'autoencodeur ne renvoie pas la même valeur de capacité à droite pour tous les fils, nous devons en faire la moyenne ici pour conserver la cohérence par rapport à l'image et ne pas en favoriser une partie.
        # On appelle récursivement la méthode dessiner sur les enfants
        for i in range(nombreDivisions - 1):
            fils = self.fils[i]
            coinHautGaucheFils = (
                coinHautGaucheY + i * hauteur,
                coinHautGaucheX,
            )
            coinBasDroiteFils = (
                coinHautGaucheY + (i + 1) * hauteur,
                coinBasDroiteX,
            )
            fils.lire(
                image, coinHautGaucheFils, coinBasDroiteFils, nouveauConteneur
            )
        # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
        fils = self.fils[-1]
        coinHautGaucheFils = (
            coinHautGaucheY + (nombreDivisions - 1) * hauteur,
            coinHautGaucheX,
        )
        coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
        fils.lire(image, coinHautGaucheFils, coinBasDroiteFils, conteneur)
        # On renvoie la moyenne des valeurs des capacites proposées.
        if conteneur is None:
            # On a créé notre propre conteneur, il faut à présent le résoudre.
            return sum(nouveauConteneur) / len(nouveauConteneur)
        else:
            # L'un des parents résoudra le conteneur le moment venu
            return None

    def normaliser(self, image, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On calcule la hauteur de chaque subdivision verticale que l'on s'apprète à créer. On rappelle que le 0,0 est tout en haut à gauche de l'image.
        hauteur = int((coinBasDroiteY - coinHautGaucheY) / nombreDivisions)
        if hauteur == 0:
            raise ImageTropPetite(
                f"La hauteur d'une liaison parallèle normalisée à la profondeur {self.profondeur} serait 0, la forme de l'arbre est {self.arbre.forme}"
            )
        # Dans la mesure où l'autoencodeur ne renvoie pas la même valeur de capacité à droite pour tous les fils, nous devons en faire la moyenne ici pour conserver la cohérence par rapport à l'image et ne pas en favoriser une partie.
        # On appelle récursivement la méthode dessiner sur les enfants
        for i in range(nombreDivisions - 1):
            fils = self.fils[i]
            coinHautGaucheFils = (
                coinHautGaucheY + i * hauteur,
                coinHautGaucheX,
            )
            coinBasDroiteFils = (
                coinHautGaucheY + (i + 1) * hauteur,
                coinBasDroiteX,
            )
            fils.normaliser(image, coinHautGaucheFils, coinBasDroiteFils)
        # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
        fils = self.fils[-1]
        coinHautGaucheFils = (
            coinHautGaucheY + (nombreDivisions - 1) * hauteur,
            coinHautGaucheX,
        )
        coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
        fils.normaliser(image, coinHautGaucheFils, coinBasDroiteFils)

    def elaguerSousArbre(self, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On suppose que la hauteur allouée est supérieure à 0
        hauteurAllouee = coinBasDroiteY - coinHautGaucheY
        hauteur = int(hauteurAllouee / nombreDivisions)
        while hauteur == 0 and len(self.fils) > 2:
            index = randrange(len(self.fils))
            self.suppressionFils(index)
            hauteur = int(hauteurAllouee / nombreDivisions)

        if hauteur == 0:
            # Si on sort de la boucle et que la hauteur allouée n'est toujours pas suffisante, alors on se fait remplacer par son dernier fils. Comme on a supposé que la hauteur allouée est plus grande que 0, le problème va se poser récursivement sur le parent.
            index = randrange(len(self.fils))
            # On remarque que les deux seules valeurs possibles sont 0 et 1. Or si on tire 1, 1-1 = 0 désigne le dernier enfant et de même pour 0, 0-1 = -1 désigne bien l'autre enfant.
            dernierFils = self.fils[index - 1]
            self.suppressionFils(index)
            # On appelle récursivement l'élaguage sur le dernier fils (qui a pris notre place).
            dernierFils.elaguerSousArbre(coinHautGauche, coinBasDroite)
        else:
            # On a trouvé de la place pour tous les enfants, il faut de même qu'eux fassent tenir leurs sous arbres.
            for i in range(len(self.fils) - 1):
                fils = self.fils[i]
                coinHautGaucheFils = (
                    coinHautGaucheY + i * hauteur,
                    coinHautGaucheX,
                )
                coinBasDroiteFils = (
                    coinHautGaucheY + (i + 1) * hauteur,
                    coinBasDroiteX,
                )
                fils.elaguerSousArbre(coinHautGaucheFils, coinBasDroiteFils)
            # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
            fils = self.fils[-1]
            coinHautGaucheFils = (
                coinHautGaucheY + (nombreDivisions - 1) * hauteur,
                coinHautGaucheX,
            )
            coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
            fils.elaguerSousArbre(coinHautGaucheFils, coinBasDroiteFils)



# Ce fichier contient l'implémentation des noeuds de type série. La documentation de toutes les méthodes des noeuds est disponible dans la classe parente.

from random import randrange  # Utile pour un élaguage aléatoire

from intelligenceArtificielle.structureDonnees.Noeud import (
    Noeud,
    conversionIndice,
)
from intelligenceArtificielle.structureDonnees.Coefficients import (
    capacite,
)  # Détermination aléatoire des valeurs de capacités
from intelligenceArtificielle.structureDonnees.SunRiseException import (
    NonFeuilleException,
    ImageTropPetite,
)


class Serie(Noeud):
    """docstring for Serie."""

    def __init__(self, arbre=None, parent=None):
        self.capacites = []
        # self.valeurCapaciteDefaut = 1 # Obsolete, remplacé par une propriété
        # Attacher fait référence à capacites et le constructeur parent fait référence à attacher, donc il faut forcément appeller le constructeur parent en dernier.
        super(Serie, self).__init__(arbre=arbre, parent=parent)

    # Note : les propriétés ne sont pas héritées en python...
    @property
    def profondeur(self):
        if self.parent is None:
            return 0
        elif self._profondeur is None:
            self._profondeur = (
                self.parent.profondeur + 1
            )  # On mémoïse la profondeur pour éviter des calculs inutiles
        return self._profondeur

    @property
    def valeurCapaciteDefaut(self):
        return capacite()

    def creationSimulationRecursive(self, A, B, C, gauche, droite, curseur):
        listeTemperatures = [gauche]
        for i in range(len(self.capacites)):
            curseur += 1
            listeTemperatures.append(curseur)
        listeTemperatures.append(droite)

        for i in range(len(self.fils)):
            gaucheBranche = listeTemperatures[i]
            droiteBranche = listeTemperatures[i + 1]
            result, curseur = self.fils[i].creationSimulationRecursive(
                A, B, C, gaucheBranche, droiteBranche, curseur
            )
            if result is not None:
                # Text (i.e. gauche is None) a un traitement différent
                if gaucheBranche is None:
                    A[droiteBranche, droiteBranche] += result
                    B[droiteBranche, 0] = -result
                else:
                    # Termes hors diagonale
                    A[gaucheBranche, droiteBranche] = -result
                    A[droiteBranche, gaucheBranche] = -result
                    # Termes diagonaux
                    A[droiteBranche, droiteBranche] += result
                    A[gaucheBranche, gaucheBranche] += result
            if i + 1 < len(self.fils):
                C[droiteBranche, droiteBranche] /= self.capacites[i]
        return None, curseur

    def creationMarquageRecursif(
        self, A, B, C, D, gauche, droite, curseur, conteneur, capaciteDroite
    ):
        listeTemperatures = [gauche]
        listeCapacite = []
        for i in range(len(self.capacites)):
            curseur += 1
            listeTemperatures.append(curseur)
            listeCapacite.append(self.capacites[i])
        listeTemperatures.append(droite)
        listeCapacite.append(capaciteDroite)

        delai = 0
        for i in range(len(self.fils)):
            gaucheBranche = listeTemperatures[i]
            droiteBranche = listeTemperatures[i + 1]
            capaciteBranche = listeCapacite[i]
            delaiBranche, resultBranche, curseur = self.fils[
                i
            ].creationMarquageRecursif(
                A,
                B,
                C,
                D,
                gaucheBranche,
                droiteBranche,
                curseur,
                None,
                capaciteBranche,
            )
            if resultBranche is not None:
                # Text (i.e. gauche is None) a un traitement différent
                if gaucheBranche is None:
                    A[droiteBranche, droiteBranche] += resultBranche
                    B[droiteBranche, 0] = -resultBranche
                else:
                    # Termes hors diagonale
                    A[gaucheBranche, droiteBranche] = -resultBranche
                    A[droiteBranche, gaucheBranche] = -resultBranche
                    # Termes diagonaux
                    A[droiteBranche, droiteBranche] += resultBranche
                    A[gaucheBranche, gaucheBranche] += resultBranche

            if delaiBranche is not None:
                di = conversionIndice(gaucheBranche)
                dj = conversionIndice(droiteBranche)
                D[di, dj] = delaiBranche
                D[dj, di] = delaiBranche
                delai += delaiBranche

            if i + 1 < len(self.fils):
                C[droiteBranche, droiteBranche] /= self.capacites[i]

        if conteneur is not None:
            conteneur.append(delai)

        return delai, None, curseur

    def injectionMarquage(self, E, gauche, droite, curseur, capaciteDroite):
        listeTemperatures = [gauche]
        listeCapacite = []
        for i in range(len(self.capacites)):
            curseur += 1
            listeTemperatures.append(curseur)
            listeCapacite.append(self.capacites[i])
        listeTemperatures.append(droite)
        listeCapacite.append(capaciteDroite)

        for i in range(len(self.fils)):
            gaucheBranche = listeTemperatures[i]
            droiteBranche = listeTemperatures[i + 1]
            capaciteBranche = listeCapacite[i]
            curseur = self.fils[i].injectionMarquage(
                E, gaucheBranche, droiteBranche, curseur, capaciteBranche
            )

        return curseur

    def suppressionFils(self, index):
        # On met à jour la forme de l'arbre en détachant le noeud désigné.
        self.fils[index].detacher()
        # IndexError si index est trop grand pour self.fils
        del self.fils[index]
        taille = len(self.fils)
        # Pour une liaison série il faut également supprimer la capacité intermédiaire.
        if index == taille:
            # i.e. on a supprimé le dernier élément de l'ancienne liste des fils.
            del self.capacites[-1]
        else:
            del self.capacites[index]
        # On met à jour le nombre de condensateurs dans l'arbre
        self.arbre.nbCondensateurs -= 1

        if taille < 2:
            # S'il ne reste plus qu'un seul bloc fils il prend la place du bloc serie. Il faut juste faire attention à la mémoïzation de la profondeur
            remplacant = self.fils[0]
            del self.fils[0]  # Pour éviter les problèmes avec détacher
            remplacant.detacher()  # On va le rattacher dans remplacer
            remplacant._profondeur = self.profondeur
            # Notez que remplacer appelle detacher pour nous en interne
            self.remplacer(remplacant)
            # On renvoie le noeud qui prend notre place dans l'arbre
            return remplacant
        else:
            return self

    def ajoutFils(self, nouveauFils, index=None, forme=None):
        if forme is not None:
            raise NonFeuilleException(
                "Vous ne pouvez pas changer le forme d'un noeud (non feuille) qui existe déjà avec ajoutFils !"
            )
        elif index is None:
            raise NonFeuilleException(
                "Vous devez préciser à quel index insérer le nouveau fils parmi les fils du noeud préexistant !"
            )
        taille = len(self.fils)
        # Il faut aussi aojuter une valeur de capacité intermédiaire dans le cas d'une liaison série. Attention à ne pas ajouter de capacité pour le premier fils !
        if taille != 0:
            if index == taille or index == taille + 1:
                self.capacites.append(self.valeurCapaciteDefaut)
            else:
                self.capacites.insert(index, self.valeurCapaciteDefaut)
            # On met à jour le nombre de condensateurs dans l'arbre.
            self.arbre.nbCondensateurs += 1

        # On insère le noeud dans la liste des fils.
        if index == len(self.fils):
            self.fils.append(nouveauFils)
        else:
            self.fils.insert(index, nouveauFils)
        # On met à jour la généalogie du noeud ajouté.
        nouveauFils.parent = self
        # On attache le nouveau noeud à l'arbre
        nouveauFils.attacher(self.arbre)
        # On garde notre place dans l'arbre, on se renvoie donc soi-même
        return self

    def substituerEnfants(self, listeFils):
        if len(listeFils) < 2:
            raise ValueError(
                f"Vous devez forcément donner plus de deux fils à un noeud série, mais l'argument passé pour listeFils ne comporte que {len(listeFils)} éléments."
            )
        # On commence par détacher tous nos enfants
        for fils in self.fils:
            fils.detacher()
        # Puis on change la liste des noeuds enfants.
        self.fils = listeFils
        for fils in self.fils:
            # On se note comme parent dans chacun des nouveaux fils
            fils.parent = self
            # On attache le nouveau fils à l'arbre
            fils.attacher(self.arbre)
        # En plus de cela, il faut gérer la variation du nombre de condensateurs.
        nouveauNbCondensateurs = len(self.fils) - 1
        differenceCapacites = nouveauNbCondensateurs - len(self.capacites)
        if differenceCapacites <= 0:
            # On supprime les capacités superflues
            self.capacites = self.capacites[:nouveauNbCondensateurs]
        else:
            # On ajoute des capacités jusqu'à satisfaction
            nouvellesCapacites = [
                self.valeurCapaciteDefaut for i in range(differenceCapacites)
            ]
            self.capacites += nouvellesCapacites
        # Enfin, il faut remettre à jour le nombre de capacités de l'arbre.
        self.arbre.nbCondensateurs += differenceCapacites

    # L'algorithme de copie du sous arbre est évidemment récursif
    def sousArbre(self):
        copieNoeud = Serie(None)
        for fils in self.fils:
            copieFils = fils.sousArbre()
            copieFils.parent = copieNoeud
            copieNoeud.fils.append(copieFils)
        for valeurCapacite in self.capacites:
            copieNoeud.capacites.append(valeurCapacite)
        return copieNoeud

    def attacher(self, arbre):
        if arbre is not self.arbre:
            self.arbre = arbre
            if self.profondeur + 1 == len(self.arbre.forme):
                # Comme un noeud série a forcément des enfants, il faut ajouter un niveau à la forme de l'arbre
                self.arbre.forme.append(0)
            self.arbre.forme[self.profondeur] += 1
            # Il faut aussi prendre en compte toutes les capacités que l'on ajoute à l'arbre.
            self.arbre.nbCondensateurs += len(self.capacites)
            for fils in self.fils:
                fils.attacher(arbre)

    def detacher(self, perdreParent=True):
        self.arbre.forme[self.profondeur] -= 1
        # On détache récursivement tous les fils pour bien prendre en compte l'influence sur la forme.
        for fils in self.fils:
            fils.detacher(perdreParent=False)
        # On décompte toutes les capacités de ce noeud.
        self.arbre.nbCondensateurs -= len(self.capacites)
        if self.arbre.forme[self.profondeur + 1] == 0:
            # i.e. on a détaché tous les noeuds qui étaient à la profondeur suivante (qui est par construction la dernière de l'arbre), alors il faut l'enlever de la forme de l'arbre pour éviter les 0 inutiles
            self.arbre.forme.pop()
        # On ne perd la référence à l'arbre qu'à la toute fin de la fonction car on la référence au dessus
        self.arbre = None
        # On perd aussi la mémoïzation de la profondeur pour éviter de la réutiliser si on est à l'intérieur d'un sous arbre qui est déplacé
        self._profondeur = None
        if perdreParent:
            # On perd aussi la référence à son parent pour éviter les effets de bord étranges
            self.parent = None

    def dessiner(self, image, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On calcule la largeur de chaque subdivision horizontale que l'on s'apprète à créer. On rappelle que le 0,0 est tout en haut à gauche de l'image.
        largeur = int((coinBasDroiteX - coinHautGaucheX) / nombreDivisions)
        if largeur == 0:
            raise ImageTropPetite(
                f"La largeur d'une liaison série écrite à la profondeur {self.profondeur} serait 0, la forme de l'arbre est {self.arbre.forme}"
            )
        # On appelle récursivement la méthode dessiner sur les enfants
        for i in range(nombreDivisions - 1):
            fils = self.fils[i]
            coinHautGaucheFils = (
                coinHautGaucheY,
                coinHautGaucheX + i * largeur,
            )
            coinBasDroiteFils = (
                coinBasDroiteY,
                coinHautGaucheX + (i + 1) * largeur,
            )
            fils.dessiner(image, coinHautGaucheFils, coinBasDroiteFils)
        # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
        fils = self.fils[-1]
        coinHautGaucheFils = (
            coinHautGaucheY,
            coinHautGaucheX + (nombreDivisions - 1) * largeur,
        )
        coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
        fils.dessiner(image, coinHautGaucheFils, coinBasDroiteFils)

    def lire(self, image, coinHautGauche, coinBasDroite, conteneur):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On calcule la largeur de chaque subdivision horizontale que l'on s'apprète à créer. On rappelle que le 0,0 est tout en haut à gauche de l'image.
        largeur = int((coinBasDroiteX - coinHautGaucheX) / nombreDivisions)
        if largeur == 0:
            raise ImageTropPetite(
                f"La largeur d'une liaison série écrite à la profondeur {self.profondeur} serait 0, la forme de l'arbre est {self.arbre.forme}"
            )
        # On appelle récursivement la méthode dessiner sur les enfants
        for i in range(nombreDivisions - 1):
            fils = self.fils[i]
            coinHautGaucheFils = (
                coinHautGaucheY,
                coinHautGaucheX + i * largeur,
            )
            coinBasDroiteFils = (
                coinBasDroiteY,
                coinHautGaucheX + (i + 1) * largeur,
            )
            # On met à jour la capacité à droite du fils avec la valeur renvoyée par lire
            self.capacites[i] = fils.lire(
                image, coinHautGaucheFils, coinBasDroiteFils, conteneur=None
            )
        # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
        fils = self.fils[-1]
        coinHautGaucheFils = (
            coinHautGaucheY,
            coinHautGaucheX + (nombreDivisions - 1) * largeur,
        )
        coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
        capaciteDroite = fils.lire(
            image, coinHautGaucheFils, coinBasDroiteFils, conteneur=None
        )
        if conteneur is not None:
            conteneur.append(capaciteDroite)
        return capaciteDroite

    def normaliser(self, image, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On calcule la largeur de chaque subdivision horizontale que l'on s'apprète à créer. On rappelle que le 0,0 est tout en haut à gauche de l'image.
        largeur = int((coinBasDroiteX - coinHautGaucheX) / nombreDivisions)
        if largeur == 0:
            raise ImageTropPetite(
                f"La largeur d'une liaison série écrite à la profondeur {self.profondeur} serait 0, la forme de l'arbre est {self.arbre.forme}"
            )
        # On appelle récursivement la méthode dessiner sur les enfants
        for i in range(nombreDivisions - 1):
            fils = self.fils[i]
            coinHautGaucheFils = (
                coinHautGaucheY,
                coinHautGaucheX + i * largeur,
            )
            coinBasDroiteFils = (
                coinBasDroiteY,
                coinHautGaucheX + (i + 1) * largeur,
            )
            # On met à jour la capacité à droite du fils avec la valeur renvoyée par lire
            fils.normaliser(image, coinHautGaucheFils, coinBasDroiteFils)
        # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
        fils = self.fils[-1]
        coinHautGaucheFils = (
            coinHautGaucheY,
            coinHautGaucheX + (nombreDivisions - 1) * largeur,
        )
        coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
        fils.normaliser(image, coinHautGaucheFils, coinBasDroiteFils)

    def elaguerSousArbre(self, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        # On compte le nombre de divisions qu'il faut faire sur l'image
        nombreDivisions = len(self.fils)
        # On suppose que la largeur allouée est supérieure à 0
        largeurAllouee = coinBasDroiteX - coinHautGaucheX
        largeur = int(largeurAllouee / nombreDivisions)
        while largeur == 0 and len(self.fils) > 2:
            index = randrange(len(self.fils))
            self.suppressionFils(index)
            largeur = int(largeurAllouee / nombreDivisions)

        if largeur == 0:
            # Si on sort de la boucle et que la largeur allouée n'est toujours pas suffisante, alors on se fait remplacer par son dernier fils. Comme on a supposé que la largeur allouée est plus grande que 0, le problème va se poser récursivement sur le dernier fils.
            index = randrange(len(self.fils))
            # On remarque que les deux seules valeurs possibles sont 0 et 1. Or si on tire 1, 1-1 = 0 désigne le dernier enfant et de même pour 0, 0-1 = -1 désigne bien l'autre enfant.
            dernierFils = self.fils[index - 1]
            self.suppressionFils(index)
            # On appelle récursivement l'élaguage sur le dernier fils (qui a pris notre place).
            dernierFils.elaguerSousArbre(coinHautGauche, coinBasDroite)
        else:
            # On a trouvé de la place pour tous les enfants, il faut de même qu'eux fassent tenir leurs sous arbres.
            for i in range(len(self.fils) - 1):
                fils = self.fils[i]
                coinHautGaucheFils = (
                    coinHautGaucheY,
                    coinHautGaucheX + i * largeur,
                )
                coinBasDroiteFils = (
                    coinBasDroiteY,
                    coinHautGaucheX + (i + 1) * largeur,
                )
                # On met à jour la capacité à droite du fils avec la valeur renvoyée par lire
                fils.elaguerSousArbre(coinHautGaucheFils, coinBasDroiteFils)
            # Pour être certain de bien colorer toute l'image et de ne pas laisser des bords noirs à cause de problèmes d'arrondis, on effectue la dernière coloration séparément.
            fils = self.fils[-1]
            coinHautGaucheFils = (
                coinHautGaucheY,
                coinHautGaucheX + (nombreDivisions - 1) * largeur,
            )
            coinBasDroiteFils = (coinBasDroiteY, coinBasDroiteX)
            fils.elaguerSousArbre(coinHautGaucheFils, coinBasDroiteFils)



"""Implémentation de la page de récupération des courbes de l'utilisateur."""

import tkinter as tk
import numpy as np  # Utile pour lire les fichiers npy
import matplotlib  # Utile pour tracer les courbes

# from matplotlib import style
# from tkinter import filedialog, Text

# Configuration de matplotlib pour pouvoir intégrer les courbes dans l'interface
# graphique.
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

from intelligenceArtificielle.Outils import (
    lireTableau,
)  # Utile pour récupérer les mesures contenues dans un fichier.
from intelligenceArtificielle.structureDonnees.SunRiseException import (
    FichierIncoherent,
)  # Lorsque les différentes bases de temps des fichiers ne sont pas les mêmes


class PageLectureDonnees(tk.Frame):
    """Classe pour la page de lecture des données."""

    def __init__(self, interfaceParente):
        """Initialisation de la classe."""
        super(PageLectureDonnees, self).__init__(
            interfaceParente.fenetrePrincipale, bg="white"
        )
        self.interfaceParente = interfaceParente
        self.place(relwidth=1, relheight=1)

        # Les conteneurs de T, Tint, Text et Pint qui permettront de récupérer
        # les valeurs une fois que l'utilisateur les aura rentrées
        self.T = None
        self.Tint = np.empty(0)
        self.Text = np.empty(0)
        self.Pint = np.empty(0)

        # Le titre de la fenêtre
        self.titre = tk.Label(
            self, text="Choix des données de mesure", fg="white", bg="green"
        )
        self.titre.pack()

        # Le bouton pour choisir la température intérieure
        self.boutonTint = tk.Button(
            self,
            text="Tint",
            padx=10,
            pady=5,
            fg="white",
            bg="gray",
            command=self.ajoutFichierTint,
        )
        self.boutonTint.place(relwidth=0.2, relheight=0.1, relx=0.1, rely=0.2)

        # Le bouton pour la température extérieure
        self.boutonText = tk.Button(
            self,
            text="Text",
            fg="white",
            bg="gray",
            command=self.ajoutFichierText,
        )
        self.boutonText.place(relwidth=0.2, relheight=0.1, relx=0.1, rely=0.5)

        # Le bouton pour la température intérieure
        self.boutonPint = tk.Button(
            self,
            text="Pint",
            padx=10,
            pady=5,
            fg="white",
            bg="gray",
            command=self.ajoutFichierPint,
        )
        self.boutonPint.place(relwidth=0.2, relheight=0.1, relx=0.1, rely=0.8)

        # Le bouton de confirmation
        self.boutonConfirmation = tk.Button(
            self,
            text="Confirmer les mesures",
            fg="white",
            bg="green",
            command=self.confirmerValeurs,
        )
        self.boutonConfirmation.pack(side=tk.BOTTOM)

    # Les alias de la fonction `ajoutFichier` avec les bons arguments pour les*
    # boutons.
    def ajoutFichierTint(self):
        self.ajoutFichier("Tint", 0.1)

    def ajoutFichierText(self):
        self.ajoutFichier("Text", 0.4)

    def ajoutFichierPint(self):
        self.ajoutFichier("Pint", 0.7)

    def ajoutFichier(self, choix, hauteurRelative):
        """
        Paramètres
        ----------
        choix
            Soit `Tint`, soit `Text`, soit `Pint`.
        hauteurRelative
            Hauteur du graphe qui sera tracé.

        Exceptions lancées
        ------------------
        FichierIncoherent
            Lorsque les différentes bases de temps des fichiers ne sont pas les
            mêmes.
        ValueError
            Lorsque `choix` n'est pas une valeur acceptée.
        """
        nomFichier = tk.filedialog.askopenfilename(
            initialdir=self.interfaceParente.dossierCourant,
            title="Sélectionnez le fichier de valeurs",
            filetypes=(
                ("tableau csv", "*.csv"),
                ("tableau numpy", "*.npy"),
                ("all files", "*./"),
            ),
        )

        # On transpose les valeurs recues pour pouvoir les séparer plus
        # facilement
        T, grandeur = lireTableau(nomFichier).T

        if self.T is None:
            # Si on n'a pas encore de valeur pour `T`, on la récupère du fichier
            self.T = T
        elif (self.T != T).any():
            # Les données sont incohérentes, problème avec la base de temps.
            raise FichierIncoherent(
                f"Les données dans le fichier {nomFichier} n'ont pas la même base de temps que celles présentes dans un autre fichier enregistré plus tôt."
            )
        # Sinon tout va bien

        if choix == "Tint":
            self.Tint = grandeur
        elif choix == "Text":
            self.Text = grandeur
        elif choix == "Pint":
            self.Pint = grandeur
        else:
            raise ValueError(
                f"Les seules valeurs possible pour choix sont Tint, Text et Pint. La valeur {choix} n'est pas reconnue."
            )

        # On remplit la figure que l'on va afficher sur l'interface.
        # On commence par créer la figure qui va contenir la courbe.
        figure = Figure(figsize=(5, 5), dpi=100)

        # On créé une nouvelle courbe
        courbe = figure.add_subplot(111)
        # Enfin on trace la courbe
        courbe.plot(self.T, grandeur)

        # Il ne reste plus qu'à afficher le résultat dans un canevas dédié
        canevasCourbe = FigureCanvasTkAgg(figure, self)
        canevasCourbe.get_tk_widget().place(
            relwidth=0.7, relheight=0.25, relx=0.3, rely=hauteurRelative
        )

    def confirmerValeurs(self):
        self.interfaceParente.pageSuivante()



"""Ce fichier contient un ensemble de méthodes utiles à l'entrainement des
autoencodeurs."""

import numpy as np
import os  # Utile pour les manipulations de fichiers
from intelligenceArtificielle.GenerateurArbres import GenerateurArbres
from intelligenceArtificielle.Outils import (
    lireTableau,
    lireDossier,
)  # Utile pour récupérer les mesures


def creationPopulationLectureDossier(
    dossier, Cint, taillePopulation, largeur, hauteur
):
    """Renvoie une population créée à partir des mesures trouvées dans le
    dossier en argument. Notez que l'on n'élague pas la population créée si
    `largeur` ou `hauteur` est None.

    Paramètres
    ----------
    dossier : str
        Dossier où trouver les mesures pour créer la population.
    Cint : float
        Capacité thermique associée à l'air intérieur.
    taillePopulation : int
        La taille de la population à créer et à faire évoluer.
    largeur : int
        Largeur d'image souhaitée. Si `largeur` ou `hauteur` est None, alors la
        population ne sera pas élaguée.
    hauteur : int
        Hauteur d'image souhaitée. Si `largeur` ou `hauteur` est None, alors la
        population ne sera pas élaguée.
    """
    T, Tint, Text, Pint = lireDossier(dossier)
    if largeur is not None and hauteur is not None:
        elaguageForce = True
    else:
        elaguageForce = False

    return GenerateurArbres(
        Cint,
        T,
        Tint,
        Text,
        Pint,
        taillePopulation,
        largeur,
        hauteur,
        elaguageForce,
    ).population


def lectureBlob(Cint, taillePopulation, largeur, hauteur):
    """Renvoie la concaténation des populations créées à partir des données dans
    les dossiers de blob.

    Cette fonction doit être lancée lorsque le répertoire courant est la racine
    de ce dépôt git.

    Paramètres
    ----------
    Cint : float
        Capacité thermique associée à l'air intérieur.
    taillePopulation : int
        La taille de la population à créer et à faire évoluer.
    largeur : int
        Largeur d'image souhaitée.
    hauteur : int
        Hauteur d'image souhaitée.
    """
    listeDossiers = os.listdir(os.path.join("blob", "mesures"))
    population = []
    for dossier in listeDossiers:
        print(
            "Récupération des mesures dans le dossier",
            os.path.join("blob", "mesures", dossier),
        )
        # On ajoute la population créée à partir du dossier dans notre
        # population totale. Notez que si `largeur` ou `hauteur` est None, la
        # population ne sera pas élaguée.
        population += creationPopulationLectureDossier(
            os.path.join("blob", "mesures", dossier),
            Cint,
            taillePopulation,
            largeur,
            hauteur,
        )
    return population



"""Ce fichier contient la classe parente abstraite des noeuds qui composent nos
arbres. Presque tous les autres fichiers doivent importer celui ci.

La classe parente contient également de la documentation sur les différentes
méthodes des noeuds des arbres. Notez que la plupart des méthodes sont
implémentées dans les classes filles, d'où les NotImplementedError un peu
partout dans cette classe.
"""


class Noeud(object):
    """Classe pour les noeuds."""

    def __init__(self, arbre=None, parent=None):
        # IMPORTANT : il est presque toujours meilleur d'appeler le constructeur
        # sans argument et de laisser ajoutFils s'occuper de tout.
        super(Noeud, self).__init__()
        self.parent = parent
        self.arbre = None
        self.fils = []
        self._profondeur = None
        # IMPORTANT : il faut toujours attacher en dernier, car attacher peut
        # faire référence à d'autres attributs de la classe.
        if arbre is not None:
            self.attacher(arbre)

    def suppressionFils(self, index):
        """Méthode abstraite, on supprime le fils à l'index i.

        Le noeud est modifié en place, et on renvoie le parent obtenu (car c'est
        plus pratique pour la mutation des arbres).
        """
        raise NotImplementedError(
            "La suppression d'un fils n'a pas été réimplémenté."
        )

    def creationSimulationRecursive(self, A, B, C, gauche, droite, curseur):
        """Méthode abstraite, permet de calculer les matrices qui seront
        utilisées pour le calcul du score de l'arbre proposé."""
        raise NotImplementedError(
            "La création récursive des matrices de simulation n'a pas été implémentée."
        )

    def creationMarquageRecursif(
        self, A, B, C, D, gauche, droite, curseur, conteneur, capaciteDroite
    ):
        """Méthode abstraite, parcourt l'arbre en profondeur et calcule les
        matrices de simulation et la matrice des délais dans notre arbre afin de
        pouvoir faire le calcul de rétropropagation d'erreur."""
        raise NotImplementedError(
            "La création récursive du marquage de l'image n'a pas été réimplémentée."
        )

    def injectionMarquage(self, E, gauche, droite, curseur, capaciteDroite):
        """Méthode abstraite, parcourt l'arbre en notant sur chaque noeud les
        valeurs qui seront utilisées pour l'intensité des trois couleurs de
        l'image."""
        raise NotImplementedError(
            "L'injection récursive du marquage n'a pas été réimplémentée."
        )

    def ajoutFils(self, nouveauFils, index, forme=None):
        """Méthode abstraite, contraire de suppressionFils.

        Le noeud est modifié en place, et on renvoie le parent obtenu (car c'est
        plus pratique pour l'initialisation des arbres).
        """
        raise NotImplementedError("L'ajout d'un fils n'a pas été implémenté.")

    def substituerEnfants(self, listeFils):
        """Méthode abstraite, utile pour la fusion d'arbres. Remplace tous les
        enfants du noeud choisi par ceux passés en argument.

        Dans le cas d'un noeud série, on conserve autant de capacités que
        possible.
        """
        raise NotImplementedError(
            "La subtitution de noeuds enfants n'a pas été réimplémentée."
        )

    @property
    def profondeur(self):
        """Implémentation abstraite de la profondeur."""
        raise NotImplementedError(
            "Le calcul de la profondeur d'un noeud n'a pas été implémenté."
        )

    def remplacer(self, nouveau):
        """Remplace l'instance noeud désignée par nouveau dans son parent.

        Il s'agit d'une méthode interne à la classe noeud qui ne doit pas être
        appelée de l'extérieur.
        """
        if self.parent is not None:
            # On trouve l'indice de l'élément actuel dans la liste fils du
            # parent
            indexParent = self.parent.fils.index(self)
            self.parent.fils[indexParent] = nouveau
        elif self.arbre is not None:
            # Autre cas important à traiter : on est en train de changer la
            # racine de l'arbre
            self.arbre.racine = nouveau
        else:
            raise IndexError(
                "L'indivdu selectionné n'a pas de parent et n'est pas la racine d'un arbre, il n'y a donc nulle part où le remplacer."
            )
        nouveau.parent = self.parent
        # On attache le nouvel élément à la arbre
        nouveau.attacher(self.arbre)
        # Dans la mesure où l'on a perdu notre place dans l'arbre, il faut s'en
        # détacher.
        self.detacher()

    def sousArbre(self):
        """Méthode abstraite, utile pour la fusion d'arbres. Renvoie une copie
        de tout le sousArbre qui se trouve au dela de cet arbre, mais sans
        parent pour ce noeud et sans arbre associé.
        """
        raise NotImplementedError(
            "La création d'un sous-arbre à partir d'un noeud n'a pas été implémentée."
        )

    def attacher(self, arbre):
        """Méthode abstraite, attache un noeud et tous ses fils récursivement à
        un arbre, en mettant au passage à jour les caractéristiques de la
        arbre.

        Doit être appelé après avoir donné un parent au noeud considéré. Cette
        étape est utile pour garder des arbres cohérents après les étapes de
        fusion.
        """
        raise NotImplementedError(
            "L'attachement d'un noeud à un arbre n'a pas été implémentée."
        )

    def detacher(self, perdreParent=True):
        """Méthode abstraite, pendant de `attacher` utilisé pour la suppression.

        L'attribut perdre parent a été ajouté suite à des bugs ; en fait, si
        l'on détache tout un sous arbre on veut que seul le sommet le sommet du
        sous arbre perde le référence à son parent, mais l'intérieur de l'arbre
        doit rester cohérent. De l'extérieur, il n'y a normalement jamais besoin
        d'appeler `detacher` du tout, mais s'il le faut il est sans doute plus
        judicieux de ne pas passer d'argument.
        """
        raise NotImplementedError(
            "Le détachement d'un noeud d'un arbre n'a pas été implémenté."
        )

    def dessiner(self, image, coinHautGauche, coinBasDroite):
        """Méthode abstraite, à utiliser récursivement pour obtenir la
        représentation sous forme d'image de l'arbre."""
        raise NotImplementedError(
            "La coloration de l'image à partir d'un noeud n'a pas été réimplémenté."
        )

    def lire(self, image, coinHautGauche, coinBasDroite, conteneur):
        """Méthode abstraite, à utiliser récursivement pour mettre à jour un
        arbre à partir d'une image.

        Renvoie la valeur de la capacité à droite du noeud donné. On utilise le
        même système de conteneur que dans `creationMarquageRecursif` pour
        éviter les problèmes de non linéarité dans les calculs avec les liaisons
        parallèles.
        """
        raise NotImplementedError(
            "La mise à jour d'un arbre à partir d'une image n'a pas été réimplémenté."
        )

    def normaliser(self, image, coinHautGauche, coinBasDroite):
        """Méthode abstraite, à utiliser récursivement pour normaliser une image
        en lui donnant la forme de l'arbre souhaité."""
        raise NotImplementedError(
            "La normalisation d'une image à partir d'un arbre n'a pas été réimplémenté."
        )

    def elaguerSousArbre(self, coinHautGauche, coinBasDroite):
        """Méthode abstraite, à utiliser récursivement sur un arbre pour
        l'élaguer."""
        raise NotImplementedError(
            "L'élaguage d'un sous arbre n'a pas été réimplémenté."
        )


def conversionIndice(indice):
    """Convertit l'indice utile pour A, B et C en un indice utile pour D,
    important car Text ne se trouve pas dans A alors qu'il est dans D.

    Utile dans `Serie.creationMarquageRecursif` et `Feuille.injectionMarquage`.
    """
    if indice is None:
        return 0
    else:
        return indice + 1



"""Implémentation des feuilles de nos arbres (i.e. les résistances).

La documentation de toutes les méthodes des noeuds est disponible dans la classe
parente.
"""
from intelligenceArtificielle.structureDonnees.Noeud import (
    Noeud,
    conversionIndice,
)

# Les deux imports qui suivent sont utiles pour ajoutFils
from intelligenceArtificielle.structureDonnees.Serie import Serie
from intelligenceArtificielle.structureDonnees.Parallele import Parallele
import numpy as np  # Utile pour np.mean dans la lecture d'image
import intelligenceArtificielle.structureDonnees.Coefficients as Coefficients  # Utile pour s'assurer que les valeurs lues sont bien plus grandes que les valeurs minimales
from intelligenceArtificielle.structureDonnees.Coefficients import (
    conductance,
)  # Initialisation de coefficients H aléatoires
from intelligenceArtificielle.structureDonnees.SunRiseException import (
    FeuilleException,
    NonMarqueException,
    ImageTropPetite,
)


class Feuille(Noeud):
    """Classe pour les feuilles des arbres.

    Attributs
    ---------
    H : float
        Initialisation de coefficients `H` aléatoires.
    _marquage : (float, float, float)
        propriété.
    _profondeur : int
        propriété.
    """

    def __init__(self, arbre=None, parent=None):
        """Initialisation de la classe."""
        super(Feuille, self).__init__(arbre=arbre, parent=parent)
        self.H = conductance()
        self._marquage = (None, None, None)

    # Note : les propriétés ne sont pas héritées en python...
    @property
    def profondeur(self):
        """Profondeur de la feuille.

        La profondeur de la racine est 0, celle d'un fils est égal à la
        profondeur de son parent plus 1.
        """
        if self.parent == None:
            return 0
        else:
            # Pas de mémoïzation de la profondeur pour une feuille car elle sera
            # amenée à changer
            return self.parent.profondeur + 1

    #
    @property
    def marquage(self):
        """Utilisé pour colorer l'image, calculé à l'échelle de l'arbre par la
        fonction marquage.

        On en fait une propriété pour éviter qu'il ne soit lu plusieurs fois
        entre des générations par erreur.

        Exception levées
        ----------------
        NonMarqueException
            Lorsque la feuille désignée n'a pas encore été marquée.
        """
        if self._marquage != (None, None, None):
            result = self._marquage
            self._marquage = (None, None, None)
            return result
        else:
            raise NonMarqueException(
                "La feuille désignée n'a pas encore été marquée."
            )

    def creationSimulationRecursive(self, A, B, C, gauche, droite, curseur):
        return self.H, curseur

    def creationMarquageRecursif(
        self, A, B, C, D, gauche, droite, curseur, conteneur, capaciteDroite
    ):
        delai = capaciteDroite / self.H
        if conteneur is not None:
            conteneur.append(delai)
        return delai, self.H, curseur

    def injectionMarquage(self, E, gauche, droite, curseur, capaciteDroite):
        indiceErreur = conversionIndice(droite)
        erreur = E[droite]
        self._marquage = (self.H, capaciteDroite, erreur)
        return curseur

    def suppressionFils(self, index):
        raise FeuilleException("Une feuille n'a pas de fils !")

    def ajoutFils(self, nouveauFils, index=None, forme=None):
        if forme is None or forme not in ["parallele", "serie"]:
            raise FeuilleException(
                "Pour créer un nouveau noeud à partir d'une feuille il faut préciser la forme à donner, l'un de 'parallele' ou 'serie'."
            )
        elif index is not None:
            raise FeuilleException(
                "Pour créer un nouveau noeud à partir d'une feuille il ne faut pas préciser d'index, qui n'est utilisable que pour un noeud Parallele ou Serie."
            )
        elif forme == "parallele":
            nouveauNoeud = Parallele()
            # On remplace cette feuille par son nouveau parent dans la
            # généalogie de l'arbre. Si besoin, la taille de l'arbre sera
            # changée automatiquement.
            self.remplacer(nouveauNoeud)
            nouveauNoeud.ajoutFils(self, index=0)
            nouveauNoeud.ajoutFils(nouveauFils, index=1)
        elif forme == "serie":
            nouveauNoeud = Serie()
            # On remplace cette feuille par son nouveau parent dans la
            # généalogie de l'arbre. Si besoin, la taille de l'arbre sera
            # changée automatiquement.
            self.remplacer(nouveauNoeud)
            nouveauNoeud.ajoutFils(self, index=0)
            nouveauNoeud.ajoutFils(nouveauFils, index=1)
        # Le nouveau noeud prend notre place dans la généalogie, c'est donc lui
        # que l'on renvoie.
        return nouveauNoeud

    # Pour ajouter des fils à substituerEnfants il faut d'abord en ajouter un
    # (même un faux, peu importe) avec ajoutFils, puis remplacer correctement
    # les parents du fils obtenu.
    def substituerEnfants(self, listeFils):
        raise FeuilleException(
            "Vous ne pouvez pas substituer de fils a une feuille, car une feuille n'a pas de fils."
        )

    # L'algorithme de copie du sous arbre est évidemment récursif
    def sousArbre(self):
        copieNoeud = Feuille(None)
        copieNoeud.H = self.H
        return copieNoeud

    def attacher(self, arbre):
        if arbre is not self.arbre:
            self.arbre = arbre
            self.arbre.forme[self.profondeur] += 1
            # Une feuille n'a pas de fils donc on n'a pas besoin d'appel
            # récursif ici.

    def detacher(self, perdreParent=True):
        self.arbre.forme[self.profondeur] -= 1
        self.arbre = None
        if perdreParent:
            # On perd aussi la référence à son parent pour éviter les effets de
            # bord étranges
            self.parent = None

    def dessiner(self, image, coinHautGauche, coinBasDroite):
        # On récupère le marquage de ce noeud.
        rouge, vert, bleu = self.marquage
        # On récupère les coordonnées dont on a besoin pour colorer l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        if (
            coinBasDroiteX <= coinHautGaucheX
            or coinBasDroiteY <= coinHautGaucheY
        ):
            raise ImageTropPetite(
                f"Il n'y a pas assez de place pour écrire une feuille à la profondeur {self.profondeur}, la forme de l'arbre est {self.arbre.forme}"
            )
        # On colore d'abord les résistances
        image[
            coinHautGaucheY:coinBasDroiteY, coinHautGaucheX:coinBasDroiteX, 0
        ] = rouge
        # On colore ensuite les capacités
        image[
            coinHautGaucheY:coinBasDroiteY, coinHautGaucheX:coinBasDroiteX, 1
        ] = vert
        # On colore enfin l'erreur
        # On colore ensuite les capacités
        image[
            coinHautGaucheY:coinBasDroiteY, coinHautGaucheX:coinBasDroiteX, 2
        ] = bleu

    def lire(self, image, coinHautGauche, coinBasDroite, conteneur):
        # On récupère les coordonnées dont on a besoin pour lire l'image
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        if (
            coinBasDroiteX <= coinHautGaucheX
            or coinBasDroiteY <= coinHautGaucheY
        ):
            raise ImageTropPetite(
                f"Il n'y a pas assez de place pour lire une feuille à la profondeur {self.profondeur}, la forme de l'arbre est {self.arbre.forme}"
            )
        # On calcule la moyenne des valeurs de coefficients de transmissions sur
        # la zone dédiée et on l'affecte à cette feuille.
        self.H = max(
            np.mean(
                image[
                    coinHautGaucheY:coinBasDroiteY,
                    coinHautGaucheX:coinBasDroiteX,
                    0,
                ]
            ),
            Coefficients.minH,
        )
        # On calcule la moyenne des valeurs de capacites sur la zone dediée et
        # on la renvoie. Notez que l'on a rien à faire avec l'erreur proposée,
        # elle n'est qu'un indice en entrée pour aider l'autoencodeur à faire
        # son travail.
        capaciteDroite = max(
            np.mean(
                image[
                    coinHautGaucheY:coinBasDroiteY,
                    coinHautGaucheX:coinBasDroiteX,
                    1,
                ]
            ),
            Coefficients.minC,
        )
        if conteneur is not None:
            conteneur.append(capaciteDroite)
        return capaciteDroite

    def normaliser(self, image, coinHautGauche, coinBasDroite):
        # On récupère les coordonnées dont on a besoin pour normaliser l'image.
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        if (
            coinBasDroiteX <= coinHautGaucheX
            or coinBasDroiteY <= coinHautGaucheY
        ):
            raise ImageTropPetite(
                f"Il n'y a pas assez de place pour normaliser une feuille à la profondeur {self.profondeur}, la forme de l'arbre est {self.arbre.forme}"
            )
        # On calcule la moyenne des valeurs de coefficients de transmissions sur
        # la zone dédiée on l'affecte à toute la zone.
        image[
            coinHautGaucheY:coinBasDroiteY, coinHautGaucheX:coinBasDroiteX, 0
        ] = np.mean(
            image[
                coinHautGaucheY:coinBasDroiteY,
                coinHautGaucheX:coinBasDroiteX,
                0,
            ]
        )
        # De même on normalise pour les capacités
        image[
            coinHautGaucheY:coinBasDroiteY, coinHautGaucheX:coinBasDroiteX, 1
        ] = np.mean(
            image[
                coinHautGaucheY:coinBasDroiteY,
                coinHautGaucheX:coinBasDroiteX,
                1,
            ]
        )
        # Et enfin pour l'erreur
        image[
            coinHautGaucheY:coinBasDroiteY, coinHautGaucheX:coinBasDroiteX, 2
        ] = np.mean(
            image[
                coinHautGaucheY:coinBasDroiteY,
                coinHautGaucheX:coinBasDroiteX,
                2,
            ]
        )

    def elaguerSousArbre(self, coinHautGauche, coinBasDroite):
        # Il n'y a rien à faire ici, par hypothès (et par le travail fait sur les liaisons Serie et Parallele) on aura toujours de la place pour une Feuille sur laquelle on appelle la méthode elaguerSousArbre.
        coinHautGaucheY, coinHautGaucheX = coinHautGauche
        coinBasDroiteY, coinBasDroiteX = coinBasDroite
        if (
            coinBasDroiteX <= coinHautGaucheX
            or coinBasDroiteY <= coinHautGaucheY
        ):
            raise ImageTropPetite(
                f"Un problème est apparu dans l'implémentation de elaguerSousArbre, il n'y a pas assez de place pour mettre une feuille à la profondeur {self.profondeur}. La forme de l'arbre est {self.arbre.forme}"
            )



"""Implémentation de la page qui sert à récupérer les données de
configuration."""

import tkinter as tk
import json  # Pour lire le fichier de configuration

import intelligenceArtificielle.structureDonnees.Coefficients as Coefficients  # Pour modifier les valeurs globales des coefficients
from intelligenceArtificielle.Genetique import (
    Genetique,  # Pour modifier les paramètres de la classe Génétique
    GenerateurArbres,  # Pour modifier les constantes sur la façon dont les arbres sont créés
)


class PageLectureConfiguration(tk.Frame):
    """Classe pour la page de lecture de la configuration."""

    def __init__(self, interfaceParente):
        super(PageLectureConfiguration, self).__init__(
            interfaceParente.fenetrePrincipale, bg="white"
        )
        self.interfaceParente = interfaceParente
        self.place(relwidth=1, relheight=1)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        # self.place(relx=1, rely=1)

        # Un champ dans lequel on stockera les valeurs pour le constructeur
        # lorsque l'utilisateur les aura confirmées
        self.valeursConstructeur = None

        # On affiche en grille tous les coefficients de la simulation et leurs
        # valeurs.

        # Pour simplifier l'ajout ou la suppression d'une ligne dans le code.
        curseurLigne = 0

        # Le titre de la fenêtre
        self.titre = tk.Label(
            self, text="Configuration de la simulation", fg="white", bg="green"
        )
        self.titre.grid(row=curseurLigne, columnspan=2, sticky=tk.N)
        curseurLigne += 1

        # PREMIERE SECTION : Valeurs pour `Coefficients`
        self.muH, curseurLigne = self.creerLigneConfiguration(
            "muH", Coefficients.muH, curseurLigne
        )
        self.sigH, curseurLigne = self.creerLigneConfiguration(
            "sigH", Coefficients.sigH, curseurLigne
        )
        self.minH, curseurLigne = self.creerLigneConfiguration(
            "minH", Coefficients.minH, curseurLigne
        )
        self.muC, curseurLigne = self.creerLigneConfiguration(
            "muC", Coefficients.muC, curseurLigne
        )
        self.sigC, curseurLigne = self.creerLigneConfiguration(
            "sigC", Coefficients.sigC, curseurLigne
        )
        self.minC, curseurLigne = self.creerLigneConfiguration(
            "minC", Coefficients.minC, curseurLigne
        )
        self.referenceH, curseurLigne = self.creerLigneConfiguration(
            "referenceH", Coefficients.referenceH, curseurLigne
        )
        self.referenceC, curseurLigne = self.creerLigneConfiguration(
            "referenceC", Coefficients.referenceC, curseurLigne
        )
        self.referenceE, curseurLigne = self.creerLigneConfiguration(
            "referenceE", Coefficients.referenceE, curseurLigne
        )

        # DEUXIEME SECTION : Valeurs pour `GenerateurArbres`
        self.profondeurMaxArbre, curseurLigne = self.creerLigneConfiguration(
            "profondeurMaxArbre",
            GenerateurArbres.PROFONDEUR_MAX_ARBRE,
            curseurLigne,
        )
        self.largeurMaxArbre, curseurLigne = self.creerLigneConfiguration(
            "largeurMaxArbre", GenerateurArbres.LARGEUR_MAX_ARBRE, curseurLigne
        )
        self.biaisAlternance, curseurLigne = self.creerLigneConfiguration(
            "biaisAlternance", GenerateurArbres.BIAIS_ALTERNANCE, curseurLigne
        )

        # TROISIEME SECTION : Valeurs pour la classe `Genetique`
        self.chanceDeMutation, curseurLigne = self.creerLigneConfiguration(
            "chanceDeMutation", Genetique.CHANCE_DE_MUTATION, curseurLigne
        )
        (
            self.pourcentageConservationFort,
            curseurLigne,
        ) = self.creerLigneConfiguration(
            "pourcentageConservationFort",
            Genetique.POURCENTAGE_CONSERVATION_FORT,
            curseurLigne,
        )
        self.chanceSurvieFaible, curseurLigne = self.creerLigneConfiguration(
            "chanceSurvieFaible", Genetique.CHANCE_SURVIE_FAIBLE, curseurLigne
        )

        # QUATRIEME SECTION : Valeurs pour le constructeur de `Genetique`
        self.Cint, curseurLigne = self.creerLigneConfiguration(
            "Cint", Coefficients.muC, curseurLigne
        )
        self.taillePopulation, curseurLigne = self.creerLigneConfiguration(
            "taillePopulation", 50, curseurLigne
        )
        self.generationMax, curseurLigne = self.creerLigneConfiguration(
            "generationMax", 25, curseurLigne
        )
        self.objectif, curseurLigne = self.creerLigneConfiguration(
            "objectif", 100, curseurLigne
        )
        self.largeurImage, curseurLigne = self.creerLigneConfiguration(
            "largeurImage", 32, curseurLigne
        )
        self.hauteurImage, curseurLigne = self.creerLigneConfiguration(
            "hauteurImage", 32, curseurLigne
        )
        self.autoencodeur, curseurLigne = self.creerLigneConfiguration(
            "autoencodeur", "base_32x32", curseurLigne
        )

        # DERNIERE SECTION : Le bouton de confirmation
        self.boutonConfirmation = tk.Button(
            self,
            text="Valider la configuration",
            fg="white",
            bg="green",
            command=self.confirmerValeurs,
        )
        self.boutonConfirmation.grid(row=curseurLigne, columnspan=2)

    def creerLigneConfiguration(
        self, nomOption, valeurDefautOption, curseurLigne
    ):
        """Ajoute à l'interface une ligne pour rentrer l'option avec le nom
        passé en argument."""
        # On commence par créer le texte descriptif qui sera à gauche
        nomOption = tk.Label(self, text=nomOption, bg="white")
        # On le place dans l'image
        nomOption.grid(row=curseurLigne, column=0, sticky=tk.W)
        # On crée la zone de texte à remplir à droite
        champOption = tk.Entry(self)
        # On la place sur la fenêtre
        champOption.grid(row=curseurLigne, column=1)
        #
        if type(valeurDefautOption) is float:
            texteValeurOption = notationScienfifique(valeurDefautOption)
        else:
            texteValeurOption = str(valeurDefautOption)
        champOption.insert(0, texteValeurOption)
        curseurLigne += 1
        return champOption, curseurLigne

    def confirmerValeurs(self):
        """Récupère les valeurs dans tous les champs et essaie de les utiliser
        pour la simulation.

        Si un problème de conversion apparait, la valeur par défaut sera gardée.
        A la fin de la méthode, on récupère la liste des valeurs pour le
        constructeur de `Genetique`.
        """

        # PREMIERE SECTION : Valeurs pour `Coefficients`
        Coefficients.muH = float(self.muH.get())
        Coefficients.sigH = float(self.sigH.get())
        Coefficients.minH = float(self.minH.get())
        Coefficients.muC = float(self.muC.get())
        Coefficients.sigC = float(self.sigC.get())
        Coefficients.minC = float(self.minC.get())
        Coefficients.referenceH = float(self.referenceH.get())
        Coefficients.referenceC = float(self.referenceC.get())
        Coefficients.referenceE = float(self.referenceE.get())

        # DEUXIEME SECTION : Valeurs pour `GenerateurArbres`
        GenerateurArbres.PROFONDEUR_MAX_ARBRE = int(
            self.profondeurMaxArbre.get()
        )
        GenerateurArbres.LARGEUR_MAX_ARBRE = int(self.largeurMaxArbre.get())
        GenerateurArbres.BIAIS_ALTERNANCE = float(self.biaisAlternance.get())

        # TROISIEME SECTION : Valeurs pour la classe `Genetique`
        Genetique.CHANCE_DE_MUTATION = float(self.chanceDeMutation.get())
        Genetique.POURCENTAGE_CONSERVATION_FORT = float(
            self.pourcentageConservationFort.get()
        )
        Genetique.CHANCE_SURVIE_FAIBLE = float(self.chanceSurvieFaible.get())

        # QUATRIEME SECTION : Valeurs pour le constructeur de `Genetique`
        self.valeursConstructeur = (
            float(self.Cint.get()),
            int(self.taillePopulation.get()),
            int(self.generationMax.get()),
            float(self.objectif.get()),
            int(self.largeurImage.get()),
            int(self.hauteurImage.get()),
            self.autoencodeur.get(),
        )

        self.interfaceParente.pageSuivante()


def notationScienfifique(nombreFlottant):
    """Renvoie la notation scientifique d'un flottant, utile pour l'affichage
    des valeurs dans l'interface graphique."""
    return "{0:.2E}".format(nombreFlottant)



"""Ce fichier contient une classe permettant d'initialiser une population
d'arbres aléatoire.

Cette fonctionnalité a été séparée de la classe Génétique pour éviter des
problèmes de dépendances circulaires.
"""

import random
from intelligenceArtificielle.structureDonnees.Arbre import (
    Arbre,
    Feuille,
    Serie,
    Parallele,
    Noeud,
)


class GenerateurArbres(object):
    """Classe pour la génération d'arbres.

    Variables globales
    ------------------
    PROFONDEUR_MAX_ARBRE : int
    LARGEUR_MAX_ARBRE : int
    BIAIS_ALTERNANCE : float
        Le biais d'alternance est un biais qui pousse l'algorithme de création
        à créer des arbres avec une alternance de Serie/Parallele pour avoir des
        solutions plus intéressantes. Il doit être entre 0 (pas de biais
        d'alternance) et 0.5 (alternance forcée).

    Attributs
    ---------
    population : list
        Liste des individus de la population. Chaque élément est un individu de
        type `Arbre`.
    Cint : float
        Capacité thermique associée à l'air intérieur.
    T : list
        La liste des temps (pour `Text`, `Tint`, `Pint`).
    Text : list
        La série des températures extérieures.
    Tint : list
        La série des températures intérieures.
    Pint : list
        La série des puissances intérieures.
    taillePopulation : int
        La taille de la population à créer et à faire évoluer.
    generationMax : int
        Le nombre maximal d'itérations de l'algorithme.
    imageLargeur : int
        Dimensions des images.
    imageHauteur : int
        Dimensions des images.
    elaguageForce : bool
        Si elaguageForce vaut True, alors chaque individu créé sera élagué pour
        tenir dans les dimension d'image spécifiées en argument du constructeur.
        Notez que contrairement à Genetique, l'élaguage forcé est ici un
        paramètre, ce qui évite d'avoir deux valeurs incohérentes dans les deux
        classes.
    """

    PROFONDEUR_MAX_ARBRE = 100
    LARGEUR_MAX_ARBRE = 50
    BIAIS_ALTERNANCE = 0.2
    SILENCE = True

    def __init__(
        self,
        Cint,
        T,
        Text,
        Tint,
        Pint,
        taillePopulation=100,
        imageLargeur=32,
        imageHauteur=32,
        elaguageForce=True,
        modeGraphique=False,
    ):
        """Initialisation de la classe."""
        super(GenerateurArbres, self).__init__()
        self.population = []
        self.Cint = Cint
        self.T = T
        self.Text = Text
        self.Tint = Tint
        self.Pint = Pint
        self.taillePopulation = taillePopulation
        self.imageLargeur = imageLargeur
        self.imageHauteur = imageHauteur
        self.elaguageForce = elaguageForce
        # En mode graphique on initialise la population avec une fonction génératrice de sorte à obtenir une barre de progression.
        if not modeGraphique:
            # On initialise directement la population à partir du constructeur.
            self.populationAleatoire()

    def populationAleatoire(self):
        """Génère une population aléatoire."""
        # On remet la population à zéro au cas où.
        self.population.clear()
        if not GenerateurArbres.SILENCE:
            proportionsAffiche = []

        for i in range(self.taillePopulation):
            if not GenerateurArbres.SILENCE:
                # La génération de population est assez longue, on affiche des
                # messages pour maintenir l'utilisateur éveillé.
                proportion = 100 * i // self.taillePopulation
                if proportion % 5 == 0 and proportion not in proportionsAffiche:
                    proportionsAffiche.append(proportion)
                    print(f"{proportion}% de la population créé.")
            self.population.append(self.individuAleatoire())

    def populationAleatoireAnimee(self):
        """Génère une population aléatoire. Variante pour faire une barre de progression dans l'interface graphique."""
        # On remet la population à zéro au cas où.
        self.population.clear()
        proportionsAffiche = []

        for i in range(self.taillePopulation):
            # La génération de population est assez longue, on affiche des
            # messages pour maintenir l'utilisateur éveillé.
            proportion = 100 * i // self.taillePopulation
            if proportion % 5 == 0 and proportion not in proportionsAffiche:
                proportionsAffiche.append(proportion)
                yield proportion
            self.population.append(self.individuAleatoire())

    def individuAleatoire(self):
        """Génère un individu aléatoire.

        Notes
        -----
        graineAlternance : float
            `graineAlternance` est une graine qui détermine si l'alternance
            devrait commencer par des liaisons séries ou parallèles.
            `graineAlternance` peut prendre les valeurs
            -Genetique.BIAIS_ALTERNANCE ou Genetique.BIAIS_ALTERNANCE.
        hasard : bool
            `hasard` a 50% de chances d'être vrai et 50% d'être faux et nous
            permet de choisir entre ajouter un fils en série et en parallèle.
            `graineAlternance` permet de favoriser tantôt les liaisons séries,
            tantôt les liaisons parallèles.
        largeurAutorisee : int
            Pour éviter des configurations absurdes, on force l'arbre à avoir
            une forme un peu conique (c'est-à-dire à n'avoir pas trop de fils
            sur les premières générations).
        """
        graineAlternance = (
            GenerateurArbres.BIAIS_ALTERNANCE
            + 2 * (random.random() > 0.5) * GenerateurArbres.BIAIS_ALTERNANCE
        )

        # On initialise l'individu que l'on va créer.
        individu = Arbre(self.Cint, self.T, self.Text, self.Tint, self.Pint)
        # On tire aléatoirement la profondeur qu'aura notre individu.
        profondeur = random.randint(1, GenerateurArbres.PROFONDEUR_MAX_ARBRE)

        # On crée itérativement tous les niveaux de notre arbre.
        for i in range(profondeur - 1):
            # cf. docstring pour `largeurAutorisee`.
            # Notez que la largeur réelle d'une couche de l'arbre dépasse
            # souvent la largeur maximale autorisée. En effet, lorsque l'on
            # ajoute un fils a une feuille, cette dernière descend dans la
            # nouvelle profondeur. On se retrouve donc avec deux noeuds de plus
            # à la nouvelle profondeur au lieu d'un.
            largeurAutorisee = min(
                GenerateurArbres.LARGEUR_MAX_ARBRE, 3 * (i + 1)
            )
            # On détermine au hasard la largeur de la profondeur i+1.
            largeur = random.randint(1, largeurAutorisee)

            # On choisit (avec remise) les noeuds de la génération i qui
            # accueilleront des enfants à la génération i+1.
            indicesChoisis = random.choices(
                population=range(individu.forme[i]), k=largeur
            )

            # On ajoute les enfants aux parents désignés.
            for k in indicesChoisis:
                # Le noeud auquel on doit ajouter un enfant. Un même indice peut
                # être tiré plusieurs fois.
                parent = individu.inspecter(i, k)
                enfant = Feuille()
                if type(parent) is Feuille:
                    # cf. docstring pour `hasard`.
                    hasard = random.random() > (0.5 + graineAlternance)
                    if hasard:
                        parent.ajoutFils(enfant, forme="serie")
                    else:
                        parent.ajoutFils(enfant, forme="parallele")
                else:
                    # On tire aléatoirement l'index où l'on va insérer le nouvel
                    # enfant parmi les enfants préexistants du parent. Le +1
                    # correspond au fait que l'on peut insérer un enfant tout à
                    # la fin de la liste des descendants du parent.
                    indexAleatoire = random.randrange(len(parent.fils) + 1)
                    # On insère le nouvel enfant.
                    parent.ajoutFils(enfant, index=indexAleatoire)
            # On alterne le biais entre liaisons séries et parallèle.
            graineAlternance *= -1

        # Si on a demandé un élaguage automatique, alors on tronque l'individu
        # pour qu'il soit représentable sous forme d'image.
        if self.elaguageForce:
            individu.elaguer(
                largeur=self.imageLargeur, hauteur=self.imageHauteur
            )
        # On renvoie la liste ainsi formée.
        return individu



"""Ce fichier contient une implémentation déterministe de la classe abstraite
Autoencodeur (c'est à dire des autoencodeurs classiques).
"""
import os  # Utile pour les manipulations de fichiers.
import numpy as np  # Utile pour fournir des données à Keras.

from keras.layers import (
    Input,
    Dense,
    Conv2D,
    MaxPooling2D,
    UpSampling2D,
)  # Eléments constitutifs de notre réseau de neurones
from keras.models import (
    Model,
    load_model,
)  # Outils pour manipuler facilement des modèles
from keras import backend as K

from intelligenceArtificielle.Autoencodeur import (
    Autoencodeur,
)  # La classe que l'on cherche à implémenter


class AutoencodeurDeterministe(Autoencodeur):
    """Classe pour les autoencodeurs déterministes.

    Attributs
    ---------
    largeur : int
        Dimension des images. Utile dans plusieurs méthodes distinctes.
    hauteur : int
        Dimension des images. Utile dans plusieurs méthodes distinctes.
    autoencodeur : modèle Keras
        Modèle choisi.
    """

    def __init__(
        self,
        nomDuModele="ameliorateur",
        largeur=32,
        hauteur=32,
        *args,
        **kwargs
    ):
        """Initialisation de la classe.

        Paramètres
        ----------
        nomDuModele : str
            Nom du moèle à charger. None si l'on ne souhaite pas charger de
            modèle. On ira charger le fichier "src/modeles/nomDuModele.h5".
        largeur : int
        hauteur : int

        Exceptions levées
        -----------------
        ModeleIntrouvable
            Lorsque nomDuModèle.h5 n'a pas pu être trouvé.
        """
        super(AutoencodeurDeterministe, self).__init__(
            nomDuModele, largeur, hauteur, *args, *kwargs
        )
        # Le constructeur est hérité de Autoencodeur, rien de plus à faire ici

    def creation(
        self, facteurReduction=2, baseDimensions=9, baseNoyau=4, baseDense=50
    ):
        """Méthode à utiliser pour créer un nouveau modèle dans le contexte d'un
        entrainement.

        La meilleure configuration d'autoencodeur que j'ai trouvée pour notre
        problème. Il est concu pour des images de taille 32 x 32 mais d'autres
        configurations sont possibles.

        Paramètres
        ----------
        facteurReduction : int
            ou "pool_size". Facteur de réduction pour les convolutions du
            réseau.
        baseDimensions : int | list
            Dimensions des filtres de convolution du réseau.
        baseDense : int | list
            Dimensions pour les filtres "Dense" du réseau.
        baseNoyau : int | list
            ou "kernel_size". Dimensions des noyaux de convolution du réseau.

        Notes
        -----
        On peut donner soit une liste, soit un unique coefficient dans
        `baseDimensions`, `baseDense` et `baseNoyau`. Si on donne une liste, ce
        sont les coefficients dans cette liste qui sont utilisés.
        """

        # Pour baseDimensions
        if type(baseDimensions) is list:
            dimensions = baseDimensions[:4]
        else:
            # On a fourni un unique coefficent en argument
            dimensions = [i * baseDimensions for i in range(1, 5)]

        # De même pour baseNoyau
        if type(baseNoyau) is list:
            noyau = baseNoyau[:4]
        else:
            noyau = [(5 - i) * baseNoyau for i in range(1, 5)]

        # Et enfin pour baseDense
        if type(baseDense) is list:
            dense = baseDense[:2]
        else:
            dense = [2 * baseDense, baseDense]

        # La fonction utilisée par la librairie Keras pour construire notre
        # autoencodeur. Dans la notation Keras, la forme de entree est
        # (None, hauteur, largeur, 3).
        entree = Input(shape=(self.hauteur, self.largeur, 3))

        # x sera de taille (None, hauteur, largeur, 6)
        x = Conv2D(
            filters=dimensions[0],
            kernel_size=noyau[0],
            activation="relu",
            padding="same",
        )(entree)
        # encode sera de taille (None, hauteur/2, largeur/2, 6)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Conv2D(
            filters=dimensions[1],
            kernel_size=noyau[1],
            activation="relu",
            padding="same",
        )(x)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Conv2D(
            filters=dimensions[2],
            kernel_size=noyau[2],
            activation="relu",
            padding="same",
        )(x)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Conv2D(
            filters=dimensions[3],
            kernel_size=noyau[3],
            activation="relu",
            padding="same",
        )(x)
        x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Dense(dense[0], activation="relu")(x)
        # x = MaxPooling2D(pool_size=facteurReduction, strides=None)(x)

        x = Dense(dense[1], activation="relu")(x)

        # x = UpSampling2D(size=facteurReduction)(x)
        x = Dense(dense[0], activation="relu")(x)

        x = UpSampling2D(size=facteurReduction)(x)
        x = Conv2D(
            filters=dimensions[3],
            kernel_size=noyau[3],
            activation="relu",
            padding="same",
        )(x)

        x = UpSampling2D(size=facteurReduction)(x)
        x = Conv2D(
            filters=dimensions[2],
            kernel_size=noyau[2],
            activation="relu",
            padding="same",
        )(x)

        x = UpSampling2D(size=facteurReduction)(x)
        x = Conv2D(
            filters=dimensions[1],
            kernel_size=noyau[1],
            activation="relu",
            padding="same",
        )(x)

        # y sera de taille (None, hauteur, largeur, 6)
        x = UpSampling2D(size=facteurReduction)(x)
        # z sera de taille (None, hauteur, largeur, 6)
        x = Conv2D(
            filters=dimensions[0],
            kernel_size=noyau[0],
            activation="relu",
            padding="same",
        )(x)

        # decode sera de taille (None, hauteur, largeur, 3)
        decode = Conv2D(3, (3, 3), activation="sigmoid", padding="same")(x)

        # On compile le modèle créé.
        self.autoencodeur = Model(entree, decode)
        self.autoencodeur.compile(
            optimizer="adadelta", loss="binary_crossentropy"
        )
        # Utile pour avoir la forme de l'autoencodeur lorsque l'on debug.
        self.autoencodeur.summary()

    def ameliorerArbres(self, listeArbres, iterations=1):
        """Version plus haut niveau de `predire` qui travaille directement sur
        des arbres pour plus de simplicité dans l'algorithme génétique.

        Paramètres
        ----------
        listeArbres : list
        iterations : int
        """

        # On crée les images de tous nos arbres
        listeImages = np.array(
            [
                arbre.ecritureImage(largeur=self.largeur, hauteur=self.hauteur)
                for arbre in listeArbres
            ]
        )
        # On propose des version améliorées de toutes ces images
        listeImagesAmeliorees = self.autoencodeur.predict(listeImages)

        # On répète la boucle d'amélioration autant de fois que demandé.
        # Attention cependant, si on n'a demandé qu'une seule passe
        # d'amélioration on n'a pas besoin de rentrer dans cette boucle
        # (d'où le -1)
        for passe in range(iterations - 1):
            # On normalise toutes les images avec l'arbre correspondant.
            for i in range(len(listeArbres)):
                arbre = listeArbres[i]
                listeImagesAmeliorees[i] = arbre.normalisationImage(
                    listeImagesAmeliorees[i]
                )
            # On améliore à nouveau les images.
            listeImagesAmeliorees = self.autoencodeur.predict(
                listeImagesAmeliorees
            )

        # Enfin, on remet à jour tous les arbres en leur donnant des
        # coefficients améliorés.
        for i in range(len(listeArbres)):
            arbre = listeArbres[i]
            imageAmelioree = listeImagesAmeliorees[i]
            # ATTENTION, la méthode `lectureImage` détruit l'image passée en
            # argument.
            arbre.lectureImage(imageAmelioree)



"""Ce fichier contient une classe implémentant un autoencodeur classique.

Il s'agit en réalité d'un simple enrobage de la librairie Keras. Par défaut,
l'autoencodeur chargé sera celui situé dans le dossier modeles avec le nom
"ameliorateur.md5". Pour utiliser un autre modèle, il faut le placer dans le
dossier "modele" et passer son nom au constructeur.
"""

import os  # Utile pour les manipulations de fichiers.

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
import numpy as np  # Utile pour fournir des données à Keras.

# On importe keras pour pouvoir utiliser notre autoencodeur.
from keras.layers import Input, Dense, Conv2D, MaxPooling2D, UpSampling2D
from keras.models import Model, load_model
from keras import backend as K

from intelligenceArtificielle.structureDonnees.SunRiseException import (
    ModeleIntrouvable,
)
from intelligenceArtificielle.Entrainement import (
    lectureBlob,
)  # Utile pour entrainer une population très simplement.


class Autoencodeur(object):
    """Classe abstraite pour les autoencodeurs.

    On utilise une classe abstraite et de l'héritage pour séparer Autoencodeur
    déterministe et non déterministe.

    Attributs
    ---------
    largeur : int
        Dimension des images. Utile dans plusieurs méthodes distinctes.
    hauteur : int
        Dimension des images. Utile dans plusieurs méthodes distinctes.
    autoencodeur : modèle Keras
        Modèle choisi.
    """

    def __init__(
        self,
        nomDuModele="ameliorateur",
        largeur=32,
        hauteur=32,
        *args,
        **kwargs,
    ):
        """Initialisation de la classe.

        Paramètres
        ----------
        nomDuModele : str
            Nom du moèle à charger. None si l'on ne souhaite pas charger de
            modèle. On ira charger le fichier "src/modeles/nomDuModele.h5".
        largeur : int
        hauteur : int

        Exceptions levées
        -----------------
        ModeleIntrouvable
            Lorsque nomDuModèle.h5 n'a pas pu être trouvé.
        """
        super(Autoencodeur, self).__init__(*args, **kwargs)

        self.largeur = largeur
        self.hauteur = hauteur

        if nomDuModele is None:
            # Pour signifier que l'on souhaite créer un nouveau modèle et non en
            # utiliser un préexistant, utiliser la valeur None comme nomDuModele
            self.autoencodeur = None
        else:
            # sinon on charge le modèle demandé.
            # Pour que le package pip soit plus utilisable, on laisse aussi la possibiliter d'utiliser le chemin absolu du modèle que le souhaite utiliser. Pour différentier cela de l'autre option rapide, il faut mettre le .md5 à la fin du nom du modèle.
            if nomDuModele.endswith(".md5"):
                cheminModele = nomDuModele
            else:
                # On suppose que l'on se trouve dans la racine du dépôt git.
                cheminModele = os.path.join("blob", "modeles", f"{nomDuModele}.md5")
            if os.path.exists(cheminModele):
                self.autoencodeur = load_model(cheminModele)
            else:
                raise ModeleIntrouvable(
                    f"Le modele {nomDuModele} demandé n'a pas pu être chargé, le fichier {cheminModele} n'existe pas."
                )

    def creation(
        self, facteurReduction=2, baseDimensions=9, baseNoyau=4, baseDense=50
    ):
        """Méthode abstraite."""
        raise NotImplementedError(
            "La création d'un nouvel autoencodeur pour son entrainement n'a pas été réimplémentée."
        )

    # Cette méthode est identique pour les différents types d'autoencodeurs,
    # on peut donc l'implémenter ici.
    def entrainement(
        self,
        listeImagesEntree,
        listeImagesSortie,
        iterations,
        tailleGroupeEntrainement,
    ):
        """Méthode pour entrainer le réseau de neurones.

        Vous pouvez parfaitemement passer la même valeur en entrée et en sortie
        pour un entrainement à l'imitation.

        Paramètres
        ----------
        listeImagesEntree : list
        listeImagesSortie : list
        iterations : int
            ou "epochs". Nombre d'itérations sur la base d'entraînement.
        tailleGroupeEntrainement : int
            ou "batchSize".
        """

        if type(listeImagesEntree) is list:
            listeImagesEntree = np.array(listeImagesEntree)
        if type(listeImagesSortie) is list:
            listeImagesEntree = np.array(listeImagesSortie)
        self.autoencodeur.fit(
            x=listeImagesEntree,
            y=listeImagesSortie,
            batch_size=tailleGroupeEntrainement,
            epochs=iterations,
            verbose=2,
            validation_split=0.2,
            shuffle=True,
        )

    # Cette méthode est identique pour les différents types d'autoencodeurs,
    # on peut donc l'implémenter ici.
    def entrainementImitationBlob(
        self, Cint, taillePopulation, iterations, tailleGroupeEntrainement
    ):
        """Une méthode raccourcie pour entrainer le réseau à faire des
        imitations directement à partir des données présentes dans le dossier
        "blob/mesures".

        Paramètres
        ----------
        Cint
        taillePopulation
        iterations
        tailleGroupeEntrainement
        """

        # On commence par récupérer les données dans blob.
        # Les arbres seront élagués si possible.
        sourceArbresEntrainement = lectureBlob(
            Cint, taillePopulation, self.largeur, self.hauteur
        )

        # On récupère toutes les images de tous ces arbres.
        sourceImages = np.array(
            [
                arbre.ecritureImage(largeur=self.largeur, hauteur=self.hauteur)
                for arbre in sourceArbresEntrainement
            ]
        )
        # On entraine le réseau sur les images obtenues.
        self.entrainement(
            sourceImages, sourceImages, iterations, tailleGroupeEntrainement
        )

    def ameliorerArbres(self, listeArbres, iterations=1):
        """Méthode abstraite."""
        raise NotImplementedError(
            "La méthode d'amélioration d'une population d'arbres n'a pas été réimplémentée."
        )

    # Cette méthode est identique pour les différents types d'autoencodeurs,
    # on peut donc l'implémenter ici.
    def sauver(self, nomDuModele):
        """Sauvegarde le modèle actuel dans modèles sous le nom proposé.

        Paramètres
        ----------
        nomDuModele : str
            Nom souhaité pour la sauvegarde du modèle.
        """

        # On suppose que le programme est lancé depuis la racine du dépôt git.
        cheminModele = os.path.join("src", "modeles", f"{nomDuModele}.md5")
        self.autoencodeur.save(cheminModele)



"""Ce fichier contient la racine de l'interface graphique à laquelle toutes les
pages sont attachées."""

import os  # Utile pour récupérer le dossier courant si besoin
import tkinter as tk

from time import sleep  # Utile pour attendre que la convergence se termine.

from PageLectureDonnees import (
    PageLectureDonnees,
)  # La première page de l'interface
from PageLectureConfiguration import (
    PageLectureConfiguration,
)  # La deuxième page de l'interface
from PageAnimation import (
    PageAnimation,
)  # La troisième page de l'interface graphique
from PageResultats import (
    PageResultats,
)  # La quatrième page de l'interface graphique


class InterfaceGraphique(tk.Tk):
    """Classe pour l'interface graphique."""

    def __init__(self, dossierCourant=None):
        """Initialisation de la classe."""
        super(InterfaceGraphique, self).__init__()
        # On donne une taille par défaut à notre fenêtre pour qu'elle apparaisse
        # d'une façon plus harmonieuse
        self.geometry("600x600")

        # Le fait de mettre un niveau intermédiaire entre la fenêtre globale
        # (self) et les pages qui vont changer permet de faire des transitions
        # qui semblent subtilement plus fluides.
        self.fenetrePrincipale = tk.Frame(self)
        # On demande à la fenetrePrincipale d'occuper tout l'espace disponible.
        self.fenetrePrincipale.place(relwidth=1, relheight=1)

        if dossierCourant is None:
            # Si le dossier courant n'est pas précisé, on prend le dossier dans
            # lequel le programme est en train de s'exécuter.
            self.dossierCourant = os.getcwd()
        else:
            self.dossierCourant = dossierCourant

        # On commence sur la première page de l'interface
        self.pageCourante = PageLectureDonnees(self)

        # On stocke ici les mesures données par l'utilisateur.
        self.mesures = None

        # On stocke ici les paramètres utiles pour le constructeur de
        # `Genetique`.
        self.constructeur = None

        # On stocke ici la population obtenue à la fin de l'algorithme génétique
        self.populationFinale = None

        # On stocke ici le meilleur individu à la fin de l'interface graphique
        self.meilleurIndividu = None

    def afficher(self):
        """Lance l'interface graphique."""
        self.mainloop()

    def pageSuivante(self):
        """Comme son nom l'indique, permet de passer à la page suivante."""
        if type(self.pageCourante) is PageLectureDonnees:
            # On récupère les données de mesure
            self.mesures = (
                self.pageCourante.T,
                self.pageCourante.Tint,
                self.pageCourante.Text,
                self.pageCourante.Pint,
            )
            # On efface ce qui est dans la fenêtre
            self.pageCourante.destroy()
            # On affiche la page suivante
            self.pageCourante = PageLectureConfiguration(self)

        elif type(self.pageCourante) is PageLectureConfiguration:
            # On récupère les paramètres du constructeur de `Genetique`
            self.constructeur = self.pageCourante.valeursConstructeur
            # On efface ce qui est dans la fenêtre
            self.pageCourante.destroy()
            # On affiche la page suivante
            self.pageCourante = PageAnimation(
                self, self.mesures, self.constructeur
            )

            # L'appel à page suivante est fait dès que le thread de l'algorithme génétique se termine.

        else:
            # On récupère les valeurs intéressantes de l'algorithme génétique
            self.populationFinale = self.pageCourante.evolution.population
            self.meilleurIndividu = self.pageCourante.meilleurIndividu

            # On efface ce qui est dans la fenêtre
            self.pageCourante.destroy()
            T, Tint, Text, Pint = self.mesures
            # On passe à la dernière page de l'application
            self.pageCourante = PageResultats(
                self, self.populationFinale, self.meilleurIndividu, T, Tint
            )
            # L'application se termine lorsque l'on ferme la quatrième page.



"""Ce module contient les valeurs des coefficients utilisés.
Il est utilisé pour générer des coefficients aléatoires de façon cohérente ou
étalonner certaines valeurs.

Note : les valeurs définies ci dessous sont accessibles de l'extérieur comme
Coefficients.<nom> dans les autres modules où Coefficients est importé.

IMPORTANT
---------
En pratique, les équations différentielles ne sont solubles que si C >> H, (il
faut 1 à 2 ordres de grandeur de différence) pour un pas de temps de 1.
La raison est, il me semble, que le temps caractéristique de l'équation ne doit
pas être plus petit que le temps de discrétisation par le solveur. Notez
également que le solveur met significativement plus de temps à traiter les
équations en cas de problèmes que s'il n'y en a pas.
"""

from random import (
    gauss,
)  # Loi gaussienne utilisée pour générer des coefficients aléatoires


# Ces valeurs donnent la loi des coefficients de transmission H
muH = 5e-2  # Valeur moyenne
sigH = 10 * muH  # Ecart type
minH = 1e-2  # Valeur minimale (> 0)

# Fonction de génération de conductances aléatoires
def conductance():
    return max(gauss(muH, sigH), minH)


# Ces valeurs donnent la loi des coefficients de capacités C
muC = 5e6  # Valeur moyenne
sigC = 10 * muC  # Ecart type
minC = 1e6  # Valeur minimale (> 0)

# Bien régler les valeurs minimales pour que même dans le pire des cas on
# conserve deux ordres de grandeur de différence entre C et H accélère
# sensiblement l'algorithme.

# Fonction de génération de capacités aléatoires
def capacite():
    return max(gauss(muC, sigC), minC)


# Le "/10" dans ce qui suit vient du fait que sigmoid(1) ~= 0.7, donc pour avoir
# des couleurs bien étalées afin de rendre l'image plus imitable pour le réseau
# de neurones, il faut prendre une référence 10x plus petite (sigmoid(10) ~= 1).
referenceH = (muH + 5 * sigH) / 10
referenceC = (muC + 5 * sigC) / 10
referenceE = 100 / 10



"""Définit des exceptions utiles dans les différents fichiers."""


class FeuilleException(Exception):
    """Vous avez utilisé une syntaxe invalide sur une feuille."""

    pass


class NonFeuilleException(Exception):
    """Vous avez utilisé une syntaxe spécifique aux feuilles pour un noeud."""

    pass


class NonMarqueException(Exception):
    """On tente de récupérer le marquage (i.e. la couleur) d'un noeud qui n'a
    pas encore été marqué."""

    pass


class SimulationException(Exception):
    """Une simulation a échoué pour des raisons mathématiques."""

    pass


class ModeleIntrouvable(Exception):
    """Un modèle keras demandé ne peut être trouvé."""

    pass


class ImageTropPetite(Exception):
    """Un arbre est trop gros pour être représenté dans une image demandée."""

    pass


# Une exception levée par l'interface graphique lorsque les différents fichiers de mesure n'ont pas la même base de temps.
class FichierIncoherent(Exception):
    pass


# Une exception levée par l'interface graphique lorsque les différents fichiers de mesure n'ont pas la même base de temps.
class FichierIncoherent(Exception):
    pass



"""Ce fichier contient plusieurs fonctions utiles pour rendre le code de
l'interface graphique plus simple et lisible."""

import os  # Utile pour les manipulations de noms de fichiers
import numpy as np  # Utile pour les manipulations de tableaux


def lireTableau(nomDuFichier):
    """Lit un tableau et le renvoie sous la forme d'un np.array.

    Paramètres
    ----------
    nomDuFichier : string
        Tableau à lire.

    Exceptions levées
    -----------------
    ValueError
        Lorsque le fichier `nomDuFichier` n'existe pas.
    """
    baseNom, extension = os.path.splitext(nomDuFichier)
    if extension == ".npy":
        return np.load(nomDuFichier)
    elif extension == ".csv":
        return np.loadtxt(nomDuFichier, dtype=float, delimiter=",")
    else:
        raise ValueError(
            "Le fichier présenté ne ressemble à aucun type de fichiers connus."
        )


def lireDossier(nomDuDossier):
    """Simplifie l'entrainement en récupérant directement toutes les grandeurs
    d'intérêt à partir des fichiers .csv contenus dans un dossier.

    Paramètres
    ----------
    nomDuDossier : str
        Dossier cible.

    Exceptions levées
    -----------------
    ValueError
        Lorsque les fichiers (Tint.csv, Text.csv et Pint.csv) attendus
        n'existent pas dans le dossier `nomDuDossier`.
    ValueError
        Lorsque les différentes grandeurs données n'ont pas la même base de
        temps, et ne peuvent donc pas être utilisées ensemble.
    """
    fichierTint = os.path.join(nomDuDossier, "Tint.csv")
    fichierText = os.path.join(nomDuDossier, "Text.csv")
    fichierPint = os.path.join(nomDuDossier, "Pint.csv")

    # On lit les différents fichiers trouvés et on les transpose pour plus de
    # facilité.
    tableauTint = lireTableau(fichierTint).T
    tableauText = lireTableau(fichierText).T
    tableauPint = lireTableau(fichierPint).T

    # On vérifie que les données ont toutes la même base de temps.
    tempsTint, Tint = tableauTint[0], tableauTint[1]
    tempsText, Text = tableauText[0], tableauText[1]
    tempsPint, Pint = tableauPint[0], tableauPint[1]

    if (tempsTint - tempsText).any() or (tempsPint - tempsText).any():
        raise ValueError(
            "Les différentes grandeurs données n'ont pas la même base de temps et ne peuvent donc pas être utilisées ensemble."
        )
    # si tout a fonctionné, on renvoie les tableaux d'intérêt
    return tempsTint, Tint, Text, Pint



"""Implémentation de la troisième page de l'interface graphique."""

import numpy as np  # Utile pour estimer la taille à donner à la fenêtre de l'animation.
import threading  # Utile pour lancer l'optimisation sans geler l'interface graphique
import tkinter as tk
import tkinter.ttk as ttk  # Contient le widget pour les barres de progression.
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as anim

# Configuration de matplotlib pour pouvoir intégrer les courbes dans l'interface
# graphique.
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

from intelligenceArtificielle.Genetique import (
    Genetique,
    GenerateurArbres,
)  # On va réaliser l'algorithme génétique dans cette page.


# Notez que PageAnimation hérite de deux classes. Le threading.thread sert à créer un thread supplémentaire qui fera tourner l'algorithme génétique. Il s'agit d'une solution un peu étrange mais économe en code.
class PageAnimation(tk.Frame, threading.Thread):
    """Classe pour la page Animation."""

    def __init__(self, interfaceParente, donneesMesures, argumentsConstructeur):
        """Initialisation de la classe."""
        super(PageAnimation, self).__init__(
            master=interfaceParente.fenetrePrincipale, bg="white"
        )
        # Visiblement Tkinter ne support pas l'héritage multiple sur les objets de type Frame, ce qui m'oblige à utiliser cette syntaxe plus que douteuse pour initialiser le Thread (obligatoire pour lancer self.start)
        threading.Thread.__init__(self)
        self.interfaceParente = interfaceParente
        self.place(relwidth=1, relheight=1)

        # On commence par initialiser l'objet `Genetique` qui va servir à faire
        # l'algorithme. On commence par mettre certaines variables globales aux
        # bonnes valeurs.
        Genetique.SILENCE = True
        GenerateurArbres.SILENCE = True

        # Notez que le fait de mettre des 'self' partout pour récupérer ces arguments n'est pas une très bonne pratique, mais ici cela simplifie le travail pour l'écriture de la fonction 'run'.

        # On récupère les données de mesure
        T, Tint, Text, Pint = donneesMesures

        # On garde le meilleur individu en mémoire pour pouvoir l'utiliser dans la page suivante.
        self.meilleurIndividu = None

        # On récupère les paramètres du constructeur
        (
            Cint,
            taillePopulation,
            generationMax,
            objectif,
            self.imageLargeur,
            self.imageHauteur,
            autoencodeur,
        ) = argumentsConstructeur

        # Keras n'est pas thread safe, donc on est obligé d'initialiser le modèle dans le thread principal, ce qui cause un délai sensible dans le chargement de la troisième.
        self.evolution = Genetique(
            Cint,
            T,
            Text,
            Tint,
            Pint,
            taillePopulation,
            generationMax,
            objectif,
            self.imageLargeur,
            self.imageHauteur,
            autoencodeur,
            modeGraphique=True,
        )

        # On donne un titre à notre page
        self.titrePage = tk.Label(
            self, text="Réalisation de la simulation", fg="white", bg="green"
        )
        self.titrePage.pack()

        # On donne une information sur la barre de progression
        self.explicationBarre = tk.Label(
            self, text="Progression de l'initialisation", fg="black", bg="white"
        )
        self.explicationBarre.pack(pady=10)

        # On crée une barre de progression qui servira à suivre la progression
        # de l'initialisation.
        self.barreProgressionInitialisation = ttk.Progressbar(
            self, orient=tk.HORIZONTAL, length=100, mode="determinate"
        )
        self.barreProgressionInitialisation.pack(
            expand=True, fill="both", padx=10, pady=10
        )

        # On créé la figure qui contiendra l'animation principale de progression
        # de l'algorithme.
        self.figureAnimee = Figure(figsize=(5, 5), dpi=100)
        self.figureAnimee.subplots_adjust(hspace=0.5)

        # On fixe à l'avance les limites de l'animation.
        limitesAbscisse = (0, generationMax)
        limitesOrdonnee = (estimationLimitesGraphe(Tint), 0)

        # On créé deux nouvelles courbes : une pour la progression de la
        # convergence et une pour l'image du meilleur individu
        self.courbeAnimee = self.figureAnimee.add_subplot(
            211, xlim=limitesAbscisse, ylim=limitesOrdonnee
        )
        self.courbeAnimee.set_title("Convergence de l'algorithme")

        # On fixe à l'avance les axes pour les deux images
        self.imageMeilleur = self.figureAnimee.add_subplot(
            212,
            xlim=(0, self.imageHauteur - 1),
            ylim=(0, self.imageLargeur - 1),
        )
        self.imageMeilleur.set_title("Meilleur individu")

        # Les conteneurs des abscisses et ordonnées des points de l'animation
        self.X, self.Y = [], []
        (self.ligne,) = self.courbeAnimee.plot(self.X, self.Y)

        # De même, on trace le graphe pour les images
        self.matrice = np.zeros((self.imageHauteur, self.imageLargeur))
        self.image = self.imageMeilleur.imshow(self.matrice)

        # Il ne reste plus qu'à afficher le résultat dans un canevas dédié
        self.canevasCourbe = FigureCanvasTkAgg(self.figureAnimee, self)
        self.canevasCourbe.get_tk_widget().pack(side=tk.BOTTOM)

        # On lance l'algorithme génétique dans un autre thread
        self.start()

    # Le nom de cette méthode est imposée par le module threading, c'est celle qui sera lancée en parallèle avec le self.start
    def run(self):
        # On initialise la population en utilisant la méthode dédiée
        # `populationAleatoireAnimee`
        self.generateurAnimeInitialisation = (
            self.evolution.generateur.populationAleatoireAnimee()
        )

        # Tant que l'itérateur produit des valeurs.
        while True:
            try:
                self.barreProgressionInitialisation["value"] = next(
                    self.generateurAnimeInitialisation
                )
                self.update_idletasks()
            except StopIteration:
                # On a fini d'initialiser notre population, le générateur est
                # épuisé.
                self.barreProgressionInitialisation["value"] = 100
                break

        # On lance ensuite l'algorithme d'optimisation
        self.generateurOptimisation = self.evolution.optimisation()

        while True:
            try:
                argumentAnimation = next(self.generateurOptimisation)
                self.fonctionAnimation(argumentAnimation)
                # Ici on force les courbes à être retracées sur le thread
                # principal au lieu du thread de l'interface, ce qui n'est pas
                # une très bonne solution, mais je ne vois pas trop comment
                # faire autrement.
                self.canevasCourbe.draw_idle()
            except StopIteration:
                break

        # Une fois l'algorithme terminé, on passe à la page suivante
        self.interfaceParente.pageSuivante()

    def fonctionAnimation(self, argumentsAlgorithmeGenetique):
        """Fonction d'amination.

        Paramètres
        ----------
        argumentsAlgorithmeGenetique : (int, Arbre, float)
            Contient pour chaque génération `generation` le numéro de cette
            génération, son `Arbre` le plus performant et son score.

        Renvoie
        -------
        ligne : courbe
            Courbe à animer au cours du temps.
        image : image
            Image à animer au cours du temps.
        """
        (
            generation,
            self.meilleurIndividu,
            meilleurScore,
        ) = argumentsAlgorithmeGenetique
        self.X.append(generation)
        self.Y.append(meilleurScore)
        self.matrice = self.meilleurIndividu.ecritureImage(
            largeur=self.imageLargeur, hauteur=self.imageHauteur
        )
        self.ligne.set_data(self.X, self.Y)
        self.image.set_data(self.matrice)
        return self.ligne, self.image


def estimationLimitesGraphe(Tint):
    """Estime les limites à donner au graphe utilisé pour l'animation."""
    return -3 * np.mean(np.square(Tint), 0)



"""Implémentation de l'algorithme génétique."""
from operator import itemgetter  # Pour trier des listes
import random  # Pour introduire du hasard

from intelligenceArtificielle.structureDonnees.Arbre import (
    Arbre,
    Noeud,
    Feuille,
    Parallele,
    Serie,
)
from intelligenceArtificielle.GenerateurArbres import (
    GenerateurArbres,
)  # Utilisé en interne pour créer une population d'arbres aléatoires.
from intelligenceArtificielle.Autoencodeur import (
    Autoencodeur,
)  # L'interface permettant de manipuler les autoencodeurs
from intelligenceArtificielle.AutoencodeurDeterministe import (
    AutoencodeurDeterministe,
)  # Pour pouvoir instancier des autoencodeurs déterministes à partir de leurs noms.


class Genetique(object):
    """Classe pour l'algorithme génétique.

    Variables globales
    ------------------
    PROFONDEUR_MAX_ARBRE : int
    LARGEUR_MAX_ARBRE : int
    CHANCE_DE_MUTATION : float
    POURCENTAGE_CONSERVATION_FORT : float
        Pourcentage de la population qui va être gardé la génération suivante,
        sélectionne les meilleurs individus.
    CHANCE_SURVIE_FAIBLE : float
        Pourcentage de la population qui va être gardé la génération suivante,
        bien que ne figurant pas parmi les meilleurs individus.
    BIAIS_ALTERNANCE : float
        Le biais d'alternance est un biais qui pousse l'algorithme de création
        à créer des arbres avec une alternance de Serie/Parallele pour avoir des
        solutions plus intéressantes. Il doit être entre 0 (pas de biais
        d'alternance) et 0.5 (alternance forcée).
    ELAGUAGE_FORCE : bool
        Si ELAGUAGE_FORCE vaut True, alors chaque individu créé sera élagué pour
        tenir dans les dimension d'image spécifiées en argument du constructeur.
    SILENCE : bool
        Permet de lancer l'algorithme sans afficher d'information de debug dans
        la console.

    Attributs
    ---------
    population : list
        Liste des individus de la population. Chaque élément est un individu de
        type `Arbre`.
    Cint : float
        Capacité thermique associée à l'air intérieur.
    T : list
        La liste des temps (pour `Text`, `Tint`, `Pint`).
    Text : list
        La série des températures extérieures.
    Tint : list
        La série des températures intérieures.
    Pint : list
        La série des puissances intérieures.
    objectif : int
        L'objectif est le score de la fitness function à partir duquel on cesse
        de faire évoluer la population.
    taillePopulation : int
        La taille de la population à créer et à faire évoluer.
    generationMax : int
        Le nombre maximal d'itérations de l'algorithme.
    imageLargeur : int
        Dimensions des images.
    imageHauteur : int
        Dimensions des images.
    autoencodeur : AutoencodeurDeterministe
        Autoencodeur déterministe utilisé pour l'amélioration des arbres.
    """

    # Variables globales
    CHANCE_DE_MUTATION = 0.1
    POURCENTAGE_CONSERVATION_FORT = 0.5
    CHANCE_SURVIE_FAIBLE = 0.05
    ELAGUAGE_FORCE = True
    SILENCE = False

    def __init__(
        self,
        Cint,
        T,
        Text,
        Tint,
        Pint,
        taillePopulation=100,
        generationMax=1000,
        objectif=10,
        imageLargeur=32,
        imageHauteur=32,
        autoencodeur=None,
        modeGraphique=False,
    ):
        """Initialisation de la classe."""
        super(Genetique, self).__init__()
        self.Cint = Cint
        self.T = T
        self.Text = Text
        self.Tint = Tint
        self.Pint = Pint
        self.objectif = objectif
        self.taillePopulation = taillePopulation
        self.generationMax = generationMax
        self.imageLargeur = imageLargeur
        self.imageHauteur = imageHauteur
        # Si autoencodeur est None pendant l'algorithme génétique, alors la
        # phase d'amélioration des coefficients sera ignorée.
        if autoencodeur is None:
            self.autoencodeur = None
        elif type(autoencodeur) is str:
            # On peut aussi donner le nom d'un autoencodeur pour qu'il soit
            # récupéré depuis un fichier
            self.autoencodeur = AutoencodeurDeterministe(
                nomDuModele=autoencodeur,
                largeur=imageLargeur,
                hauteur=imageHauteur,
            )
        else:
            # On récupère l'autoencodeur directement de l'instance qui a été
            # fournie
            self.autoencodeur = autoencodeur
        # GénérateurArbres va automatiquement générer une population convenable
        # lors de sa construction sauf en mode graphique.
        self.generateur = GenerateurArbres(
            self.Cint,
            self.T,
            self.Text,
            self.Tint,
            self.Pint,
            self.taillePopulation,
            self.imageLargeur,
            self.imageHauteur,
            Genetique.ELAGUAGE_FORCE,
            modeGraphique,
        )
        # En mode graphique la population est initialisée plus tard avec la fonction génératrice de self.generateur (nommée populationAleatoireAnimee). Hors mode graphique, on récupère la population déjà initialisée.
        self.population = self.generateur.population

    def scorePopulation(self):
        """Calcule le score de la population.

        Renvoie une liste de couples (individu, score) triée par score sur toute
        la population. Les individus avec le meilleur score sont à la fin.
        """
        scoreIndividus = []
        if not Genetique.SILENCE:
            compteur = 0
            # Pour ne pas afficher le même message en boucle
            proportionsAffiche = []

        for individu in self.population:
            if not Genetique.SILENCE:
                proportion = 100 * compteur // self.taillePopulation
                if proportion % 5 == 0 and not proportion in proportionsAffiche:
                    proportionsAffiche.append(proportion)
                    print(f"{proportion}% de la population évalué.")
                compteur += 1
            scoreIndividus.append((individu, individu.score()))

        # On trie la liste selon l'argument d'indice 1, d'où le itemgetter.
        return sorted(scoreIndividus, key=itemgetter(1))

    def selection(self):
        """Ne conserve que les meilleurs individus de la population.

        Notes
        -----
        faiblesSurvivants
            Afin d'éviter de tomber dans des minimas locaux trop rapidement, on
            va faire survivre quelques individus faibles. On crée une liste
            dédiée pour que le dernier élément de la liste renvoyée à la fin de
            cette fonction soit bien le meilleur individu de la génération (ce
            qui n'est pas le cas si on fait directement un parent.append).
        """
        scoreIndividus = self.scorePopulation()

        # Le rang au delà duquel un individu dans la population est forcément
        # conservé.
        seuilSurvie = self.taillePopulation - int(
            Genetique.POURCENTAGE_CONSERVATION_FORT * self.taillePopulation
        )
        # On sélectionne les meilleurs individus dans la population.
        parents = scoreIndividus[seuilSurvie:]

        # cf. docstring pour `faiblesSurvivants`.
        faiblesSurvivants = []
        for individu in scoreIndividus[:seuilSurvie]:
            if random.random() < Genetique.CHANCE_SURVIE_FAIBLE:
                faiblesSurvivants.append(individu)

        parents = faiblesSurvivants + parents
        return [list(i) for i in zip(*parents)]

    def mutation(self):
        """Affecte des mutations à certains individus de la population.

        Pour l'instant, on utilise un taux de mutation constant.
        """
        for individu in self.population:
            # On note si l'individu a été muté, auquel cas il faudra l'élaguer
            # si cela est demandé pour qu'il reste représentable sous forme
            # d'image.
            besoinElaguage = False
            # On peut muter un même individu plusieurs fois si le hasard le veut
            while random.random() < Genetique.CHANCE_DE_MUTATION:
                besoinElaguage = True
                profondeur = len(individu.forme)
                # On tire une profondeur à laquelle faire la mutation.
                choixProfondeur = random.randrange(profondeur)
                # Le nombre de noeuds à la profondeur donnée.
                largeur = individu.forme[choixProfondeur]
                # On choisit un individu au hasard parmi ceux à cette profondeur.
                choixLargeur = random.randrange(largeur)
                # On récupère l'individu tiré au hasard.
                noeud = individu.inspecter(choixProfondeur, choixLargeur)

                # On tire aléatoirement entre True et False avec équiprobabilité.
                action = random.random() > 0.5

                # L'action sera différente si le noeud tiré est une `Feuille` ou
                # non.
                if type(noeud) is Feuille:
                    # On va ajouter une feuille à ce noeud.
                    fils = Feuille()
                    if action:
                        # 50% de chances d'ajouter en série
                        noeud.ajoutFils(fils, forme="serie")
                    else:
                        # 50% de chances d'ajouter en parallèle
                        noeud.ajoutFils(fils, forme="parallele")
                else:
                    nombreFils = len(noeud.fils)

                    if action:
                        # 50% de chances d'ajouter un enfant.
                        # On tire un indice sur le lequel faire notre action
                        # dans 0, `nombreFils` inclu.
                        choixIndex = random.randrange(nombreFils + 1)
                        fils = Feuille()
                        noeud.ajoutFils(fils, index=choixIndex)
                    else:
                        # 50% de chances de supprimer un fils.
                        # On tire un indice sur le lequel faire notre action
                        # dans 0, `nombreFils` exclu.
                        choixIndex = random.randrange(nombreFils)
                        noeud.suppressionFils(index=choixIndex)

            if besoinElaguage:
                # On remet l'indicateur à 0 pour le prochain individu.
                besoinElaguage = False
                if Genetique.ELAGUAGE_FORCE:
                    # Si besoin, on élague l'individu obtenu
                    individu.elaguer(
                        largeur=self.imageLargeur, hauteur=self.imageHauteur
                    )

    def fusion(self):
        """Fusionne les individus.

        On ajoutera tous les enfants d'un seul coup à la fin pour éviter de
        faire une fusion sur un enfant qui vient d'être créé.
        """
        enfants = []
        # On calcule le nombre d'individus qu'il nous reste à produire pour
        # compléter la génération suivante.
        populationManquante = self.taillePopulation - len(self.population)

        for i in range(populationManquante):
            # On tire deux individus dans la population sans faire de remise.
            pere, mere = random.sample(self.population, k=2)

            # Il nous faut à présent déterminer la profondeur maximale jusqu'à
            # laquelle les deux parents ont une forme commune.
            profondeurSemblable = 0
            # On teste toutes les profondeurs et on s'arrète à la première non
            # commune.
            for j in range(min(len(pere.forme), len(mere.forme))):
                # print(f"{j} {pere.forme[j]} {mere.forme[j]}")
                if pere.forme[j] != mere.forme[j]:
                    if j == 0:
                        print(pere.forme)
                        print(mere.forme)
                    profondeurSemblable = j - 1
                    break

            # Le sommet qui va nous servir à créer le nouvel enfant
            sommet = None
            # On tire au hasard le parent dont on va récupérer le sommet.
            choixSommet = random.random() > 0.5
            if choixSommet:
                sommet = mere.racine.sousArbre()
            else:
                sommet = pere.racine.sousArbre()

            # On crée l'objet `Arbre` qui acueille le sommet. Le constructeur
            # de `Arbre` dispose d'une syntaxe spéciale concue pour attacher
            # une racine direcetement.
            enfant = Arbre(
                self.Cint,
                self.T,
                self.Text,
                self.Tint,
                self.Pint,
                racine=sommet,
            )

            # Le nombre de noeuds qui vont devoir être fusionnés entre parent et
            # enfant.
            largeur = enfant.forme[profondeurSemblable]
            # On réalise l'étape de fusion pour tous les noeuds de
            # `profondeurSemblable`.
            for j in range(largeur):
                # On récupère le noeud correspondant pour la mere, le pere et
                # l'enfant.
                noeudPere = pere.inspecter(profondeurSemblable, j)
                noeudMere = mere.inspecter(profondeurSemblable, j)
                noeudEnfant = enfant.inspecter(profondeurSemblable, j)
                # On crée la liste de tous les nouveaux enfants possibles.
                # Un détail d'implémentation : en théorie on devrait faire un
                # choix parmi les `sousArbre` crées à partir des enfants de pere
                # et mere, mais créer un `sousArbre` peut être un peu long, donc
                # en créer qui ne seront pas utilisés est une perte de temps.
                # Au lieu de cela, on va faire notre choix sur les enfants
                # directs de pere et mere, puis calculer le `sousArbre` des
                # enfants élus, ce qui est plus efficace mais amène au même
                # résultat.
                if type(noeudEnfant) is Feuille:
                    # Si on tombe sur une feuille, on ne peut évidemment pas y
                    # ajouter de fils.
                    continue
                filsPotentiels = noeudPere.fils + noeudMere.fils
                # On détermine le nombre d'enfants à donner à l'enfant.
                nombreFils = len(filsPotentiels) // 2 + 1
                # On choisit nombreFils éléments parmi filsPotentiels sans
                # remise.
                filsChoisis = random.sample(filsPotentiels, k=nombreFils)
                # On calcule le sousArbre des enfants choisis avant de l'ajouter
                # à l'enfant.
                sousArbresChoisis = [fils.sousArbre() for fils in filsChoisis]
                # Enfin, on ajoute les nouveaux sous arbres créés sous le noeud
                # choisi de enfant.
                noeudEnfant.substituerEnfants(sousArbresChoisis)

            if Genetique.ELAGUAGE_FORCE:
                # Si besoin, on élague l'enfant obtenu pour qu'il soit
                # représentable sous forme d'image.
                enfant.elaguer(
                    largeur=self.imageLargeur, hauteur=self.imageHauteur
                )

            # On ajoute l'arbre créé à la population de descendants.
            enfants.append(enfant)
        # On renvoie la liste des enfants créés pour qu'ils soient ajoutés à la
        # population.
        return enfants

    def ameliorerPopulation(self):
        """Effectue une génération de l'algorithme génétique.

        Renvoie quelques valeurs qui sont utiles pour l'affichage."""

        # SELECTION
        if not Genetique.SILENCE:
            print("Selection")
        self.population, scores = self.selection()

        # On crée une copie du meilleur individu de la génération
        # (il risque d'être modifié par mutation).
        meilleurArbre = self.population[-1].racine.sousArbre()
        meilleurIndividu = Arbre(
            self.Cint,
            self.T,
            self.Text,
            self.Tint,
            self.Pint,
            racine=meilleurArbre,
        )

        # AMELIORATION
        if self.autoencodeur is not None:
            if not Genetique.SILENCE:
                print("Amelioration")
            self.autoencodeur.ameliorerArbres(self.population)

        # MUTATION
        if not Genetique.SILENCE:
            print("Mutation")
        self.mutation()

        # FUSION
        if not Genetique.SILENCE:
            print("Fusion")
        self.population += self.fusion()

        # On renvoie des valeurs qui pourront être utiles pour l'affichage
        # (le meilleur individu et son score).
        return meilleurIndividu, scores[-1]

    def optimisation(self):
        """Réalise l'optimisation de la population par l'algorithme génétique.

        Il s'agit d'une fonction génératrice qui "yield" les valeurs
        intéressantes pour l'animation de l'interface graphique à chaque
        génération. Une fois l'algorithme terminé, le résultat de l'algorithme
        est contenu dans les différents attributs de l'objet.
        """
        # Pour que le premier generation renvoyé soit bien 0
        generation = -1
        meilleurScore = 2 * self.objectif
        while (generation < self.generationMax) and (
            abs(meilleurScore) > self.objectif
        ):
            meilleurIndividu, meilleurScore = self.ameliorerPopulation()
            generation += 1
            yield generation, meilleurIndividu, meilleurScore
